from __future__ import print_function
import gc, os, sys
from copy import deepcopy
from warnings import warn
from time import time
import numpy as np
from scipy.sparse import coo_matrix, csc_matrix, csr_matrix

from ._LowLevelAssembly_ import _LowLevelAssembly_, _LowLevelAssemblyExplicit_, _LowLevelAssemblyLaplacian_
from ._LowLevelAssembly_ import _LowLevelAssembly_Par_, _LowLevelAssemblyExplicit_Par_

from .SparseAssemblyNative import SparseAssemblyNative
from .RHSAssemblyNative import RHSAssemblyNative

# PARALLEL PROCESSING ROUTINES
import multiprocessing
import Florence.ParallelProcessing.parmap as parmap


__all__ = ['Assemble', 'AssembleForces', 'AssembleExplicit']



def Assemble(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp):

    if fem_solver.memory_model == "shared" or fem_solver.memory_model is None:
        if not fem_solver.has_low_level_dispatcher:
            if mesh.nelem <= 600000:
                return AssemblySmall(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp)
            elif mesh.nelem > 600000:
                print("Larger than memory system. Dask on disk parallel assembly is turned on")
                return OutofCoreAssembly(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp)
        else:
            return LowLevelAssembly(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp)

    elif fem_solver.memory_model == "distributed":
        # RUN THIS PROGRAM FROM SHELL WITH python RunSession.py INSTEAD
        if not __PARALLEL__:
            warn("parallelisation is going to be turned on")

        import subprocess, os, shutil
        from time import time
        from Florence.Utils import par_unpickle
        from scipy.io import loadmat

        tmp_dir = par_unpickle(function_space,mesh,material,Eulerx,Eulerp)
        pwd = os.path.dirname(os.path.realpath(__file__))
        distributed_caller = os.path.join(pwd,"DistributedAssembly.py")

        t_dassembly = time()
        p = subprocess.Popen("time mpirun -np "+str(MP.cpu_count())+" Florence/FiniteElements/DistributedAssembly.py"+" /home/roman/tmp/",
            cwd="/home/roman/Dropbox/florence/", stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        # p = subprocess.Popen("./mpi_runner.sh", cwd="/home/roman/Dropbox/florence/", shell=True)
        p.wait()
        print('MPI took', time() - t_dassembly, 'seconds for distributed assembly')
        Dict = loadmat(os.path.join(tmp_dir,"results.mat"))

        try:
            shutil.rmtree(tmp_dir)
        except IOError:
            raise IOError("Could not delete the directory")

        return Dict['stiffness'], Dict['T'], Dict['F'], []


def LowLevelAssembly(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp):

    t_assembly = time()

    if not material.has_low_level_dispatcher:
        raise RuntimeError("Cannot dispatch to low level module since material {} does not support it".format(type(material).__name__))

    if formulation.fields == "electrostatics":
        stiffness, T = _LowLevelAssemblyLaplacian_(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp)
        fem_solver.assembly_time = time() - t_assembly
        return stiffness, T[:,None], None, None

    if fem_solver.parallel:
        stiffness, T, F, mass = ImplicitParallelLauncher(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp)
    else:
        stiffness, T, F, mass = _LowLevelAssembly_(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp)

    if isinstance(F,np.ndarray):
        F = F[:,None]
    if mass is not None:
        fem_solver.is_mass_computed = True

    fem_solver.assembly_time = time() - t_assembly

    return stiffness, T[:,None], F, mass



def AssemblySmall(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp):

    t_assembly = time()

    # GET MESH DETAILS
    C = mesh.InferPolynomialDegree() - 1
    nvar = formulation.nvar
    ndim = formulation.ndim
    nelem = mesh.nelem
    nodeperelem = mesh.elements.shape[1]

    # ALLOCATE VECTORS FOR SPARSE ASSEMBLY OF STIFFNESS MATRIX - CHANGE TYPES TO INT64 FOR DoF > 1e09
    I_stiffness=np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.int32)
    J_stiffness=np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.int32)
    # V_stiffness=np.zeros((nvar*nodeperelem)**2*nelem,dtype=np.float32)
    V_stiffness=np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.float64)

    I_mass=[]; J_mass=[]; V_mass=[]
    if fem_solver.analysis_type !='static' and fem_solver.is_mass_computed is False:
        # ALLOCATE VECTORS FOR SPARSE ASSEMBLY OF MASS MATRIX - CHANGE TYPES TO INT64 FOR DoF > 1e09
        I_mass=np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.int32)
        J_mass=np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.int32)
        V_mass=np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.float64)

    T = np.zeros((mesh.points.shape[0]*nvar,1),np.float64)
    # T = np.zeros((mesh.points.shape[0]*nvar,1),np.float32)

    mass, F = [], []
    if fem_solver.has_moving_boundary:
        F = np.zeros((mesh.points.shape[0]*nvar,1),np.float64)


    if fem_solver.parallel:
        # COMPUATE ALL LOCAL ELEMENTAL MATRICES (STIFFNESS, MASS, INTERNAL & EXTERNAL TRACTION FORCES )
        # ParallelTuple = parmap.map(formulation.GetElementalMatrices,np.arange(0,nelem,dtype=np.int32),
            # function_space, mesh, material, fem_solver, Eulerx, Eulerp)

        ParallelTuple = parmap.map(formulation,np.arange(0,nelem,dtype=np.int32),
            function_space, mesh, material, fem_solver, Eulerx, Eulerp, processes= int(multiprocessing.cpu_count()/2))

    for elem in range(nelem):

        if fem_solver.parallel:
            # UNPACK PARALLEL TUPLE VALUES
            I_stiff_elem = ParallelTuple[elem][0]; J_stiff_elem = ParallelTuple[elem][1]; V_stiff_elem = ParallelTuple[elem][2]
            t = ParallelTuple[elem][3]; f = ParallelTuple[elem][4]
            I_mass_elem = ParallelTuple[elem][5]; J_mass_elem = ParallelTuple[elem][6]; V_mass_elem = ParallelTuple[elem][6]

        else:
            # COMPUATE ALL LOCAL ELEMENTAL MATRICES (STIFFNESS, MASS, INTERNAL & EXTERNAL TRACTION FORCES )
            I_stiff_elem, J_stiff_elem, V_stiff_elem, t, f, \
            I_mass_elem, J_mass_elem, V_mass_elem = formulation.GetElementalMatrices(elem,
                function_space, mesh, material, fem_solver, Eulerx, Eulerp)

        # SPARSE ASSEMBLY - STIFFNESS MATRIX
        SparseAssemblyNative(I_stiff_elem,J_stiff_elem,V_stiff_elem,I_stiffness,J_stiffness,V_stiffness,
            elem,nvar,nodeperelem,mesh.elements)

        if fem_solver.analysis_type != 'static' and fem_solver.is_mass_computed==False:
            # SPARSE ASSEMBLY - MASS MATRIX
            SparseAssemblyNative(I_mass_elem,J_mass_elem,V_mass_elem,I_mass,J_mass,V_mass,
                elem,nvar,nodeperelem,mesh.elements)

        if fem_solver.has_moving_boundary:
            # RHS ASSEMBLY
            # for iterator in range(0,nvar):
            #     F[mesh.elements[elem,:]*nvar+iterator,0]+=f[iterator::nvar,0]
            RHSAssemblyNative(F,f,elem,nvar,nodeperelem,mesh.elements)

        # INTERNAL TRACTION FORCE ASSEMBLY
        # for iterator in range(0,nvar):
            # T[mesh.elements[elem,:]*nvar+iterator,0]+=t[iterator::nvar,0]
        RHSAssemblyNative(T,t,elem,nvar,nodeperelem,mesh.elements)


    if fem_solver.parallel:
        del ParallelTuple
        gc.collect()

    # REALLY DANGEROUS FOR MULTIPHYSICS PROBLEMS - NOTE THAT SCIPY RUNS A PRUNE ANYWAY
    # V_stiffness[np.isclose(V_stiffness,0.)] = 0.

    stiffness = coo_matrix((V_stiffness,(I_stiffness,J_stiffness)),
        shape=((nvar*mesh.points.shape[0],nvar*mesh.points.shape[0])),dtype=np.float64).tocsr()

    # stiffness = csc_matrix((V_stiffness,(I_stiffness,J_stiffness)),
        # shape=((nvar*mesh.points.shape[0],nvar*mesh.points.shape[0])),dtype=np.float32)
    # stiffness = csc_matrix((V_stiffness,(I_stiffness,J_stiffness)),
        # shape=((nvar*mesh.points.shape[0],nvar*mesh.points.shape[0])),dtype=np.float64)
    # stiffness = csr_matrix((V_stiffness,(I_stiffness,J_stiffness)),
        # shape=((nvar*mesh.points.shape[0],nvar*mesh.points.shape[0])),dtype=np.float64)

    # GET STORAGE/MEMORY DETAILS
    fem_solver.spmat = stiffness.data.nbytes/1024./1024.
    fem_solver.ijv = (I_stiffness.nbytes + J_stiffness.nbytes + V_stiffness.nbytes)/1024./1024.

    del I_stiffness, J_stiffness, V_stiffness
    gc.collect()

    if fem_solver.analysis_type != 'static' and fem_solver.is_mass_computed==False:
        mass = csr_matrix((V_mass,(I_mass,J_mass)),shape=((nvar*mesh.points.shape[0],
            nvar*mesh.points.shape[0])),dtype=np.float64)

        fem_solver.is_mass_computed = True

    fem_solver.assembly_time = time() - t_assembly

    return stiffness, T, F, mass




def OutofCoreAssembly(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp, calculate_rhs=True, filename=None, chunk_size=None):
# def OutofCoreAssembly(MainData, mesh, material, Eulerx, TotalPot, calculate_rhs=True, filename=None, chunk_size=None):
    """Assembly routine for larger than memory system of equations.
        Usage of h5py and dask allow us to store the triplets and build a sparse matrix out of
        them on disk.

        Note: The sparse matrix itfem_solver is created on the memory.
    """

    import sys, os
    from warnings import warn
    from time import time
    try:
        import psutil
    except ImportError:
        has_psutil = False
        raise ImportError("No module named psutil. Please install it using 'pip install psutil'")
    # from Core.Supplementary.dsparse.sparse import dok_matrix


    if fem_solver.parallel:
        warn("Parallel assembly cannot performed on large arrays. \n"
            "Out of core 'i.e. Dask' parallelisation is turned on instead. "
            "This is an innocuous warning")

    try:
        import h5py
    except ImportError:
        has_h5py = False
        raise ImportError('h5py is not installed. Please install it first by running "pip install h5py"')

    try:
        import dask.array as da
    except ImportError:
        has_dask = False
        raise ImportError('dask is not installed. Please install it first by running "pip install toolz && pip install dask"')

    if filename is None:
        warn("filename not given. I am going to write the output in the current directory")
        pwd = os.path.dirname(os.path.realpath(__file__))
        filename = os.path.join(pwd,"output.hdf5")


    # GET MESH DETAILS
    C = mesh.InferPolynomialDegree() - 1
    nvar = formulation.nvar
    ndim = formulation.ndim

    nelem = mesh.nelem
    nodeperelem = mesh.elements.shape[1]

    # GET MEMORY INFO
    memory = psutil.virtual_memory()
    size_of_triplets_gbytes = (mesh.points.shape[0]*nvar)**2*nelem*(4)*(3)//1024**3
    if memory.available//1024**3 > 2*size_of_triplets_gbytes:
        warn("Out of core assembly is only efficient for larger than memory "
            "system of equations. Using it on smaller matrices can be very inefficient")

    hdf_file = h5py.File(filename,'w')
    IJV_triplets = hdf_file.create_dataset("IJV_triplets",((nvar*nodeperelem)**2*nelem,3),dtype=np.float32)


    # THE I & J VECTORS OF LOCAL STIFFNESS MATRIX DO NOT CHANGE, HENCE COMPUTE THEM ONCE
    I_stiff_elem = np.repeat(np.arange(0,nvar*nodeperelem),nvar*nodeperelem,axis=0)
    J_stiff_elem = np.tile(np.arange(0,nvar*nodeperelem),nvar*nodeperelem)

    I_mass=[];J_mass=[];V_mass=[]; I_mass_elem = []; J_mass_elem = []

    if calculate_rhs is False:
        F = []
        T = []
    else:
        F = np.zeros((mesh.points.shape[0]*nvar,1),np.float64)
        T = np.zeros((mesh.points.shape[0]*nvar,1),np.float64)

    # ASSIGN OTHER NECESSARY MATRICES
    full_current_row_stiff = []; full_current_column_stiff = []; coeff_stiff = []
    full_current_row_mass = []; full_current_column_mass = []; coeff_mass = []
    mass = []

    gc.collect()

    print('Writing the triplets to disk')
    t_hdf5 = time()
    for elem in range(nelem):

        full_current_row_stiff, full_current_column_stiff, coeff_stiff, t, f, \
            full_current_row_mass, full_current_column_mass, coeff_mass = GetElementalMatrices(elem,
                MainData,mesh.elements,mesh.points,nodeperelem,Eulerx,TotalPot,I_stiff_elem,J_stiff_elem,I_mass_elem,J_mass_elem)

        IJV_triplets[(nvar*nodeperelem)**2*elem:(nvar*nodeperelem)**2*(elem+1),0] = full_current_row_stiff.flatten()
        IJV_triplets[(nvar*nodeperelem)**2*elem:(nvar*nodeperelem)**2*(elem+1),1] = full_current_column_stiff.flatten()
        IJV_triplets[(nvar*nodeperelem)**2*elem:(nvar*nodeperelem)**2*(elem+1),2] = coeff_stiff.flatten()

        if calculate_rhs is True:

            if MainData.Analysis != 'Static':
                # SPARSE ASSEMBLY - MASS MATRIX
                I_mass, J_mass, V_mass = SparseAssembly_Step_2(I_mass,J_mass,V_mass,full_current_row_mass,full_current_column_mass,coeff_mass,
                    nvar,nodeperelem,elem)

            if MainData.AssemblyParameters.ExternalLoadNature == 'Nonlinear':
                # RHS ASSEMBLY
                for iterator in range(0,nvar):
                    F[mesh.elements[elem,:]*nvar+iterator,0]+=f[iterator::nvar]
            # INTERNAL TRACTION FORCE ASSEMBLY
            for iterator in range(0,nvar):
                    T[mesh.elements[elem,:]*nvar+iterator,0]+=t[iterator::nvar,0]

        if elem % 10000 == 0:
            print("Processed ", elem, " elements")

    hdf_file.close()
    print('Finished writing the triplets to disk. Time taken', time() - t_hdf5, 'seconds')


    print('Reading the triplets back from disk')
    hdf_file = h5py.File(filename,'r')
    if chunk_size is None:
        chunk_size = mesh.points.shape[0]*nvar // 300

    print('Creating dask array from triplets')
    IJV_triplets = da.from_array(hdf_file['IJV_triplets'],chunks=(chunk_size,3))


    print('Creating the sparse matrix')
    t_sparse = time()

    stiffness = csr_matrix((IJV_triplets[:,2].astype(np.float32),
        (IJV_triplets[:,0].astype(np.int32),IJV_triplets[:,1].astype(np.int32))),
        shape=((mesh.points.shape[0]*nvar,mesh.points.shape[0]*nvar)),dtype=np.float32)

    print('Done creating the sparse matrix, time taken', time() - t_sparse)

    hdf_file.close()

    return stiffness, T, F, mass




#--------------------- ASSEMBLY ROUTINE FOR MASS MATRIX ONLY - FOR MODIFIED NEWTON RAPHSON ----------------------#
#----------------------------------------------------------------------------------------------------------------#


def AssembleMass(fem_solver, function_space, formulation, mesh, material, Eulerx):

    # GET MESH DETAILS
    C = mesh.InferPolynomialDegree() - 1
    nvar = formulation.nvar
    ndim = formulation.ndim
    nelem = mesh.nelem
    nodeperelem = mesh.elements.shape[1]

    # ALLOCATE VECTORS FOR SPARSE ASSEMBLY OF MASS MATRIX - CHANGE TYPES TO INT64 FOR DoF > 1e09
    I_mass=np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.int32)
    J_mass=np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.int32)
    V_mass=np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.float64)

    # COMPUATE ALL LOCAL ELEMENTAL MATRICES (STIFFNESS, MASS, INTERNAL & EXTERNAL TRACTION FORCES )
    ParallelTuple = parmap.map(formulation,np.arange(0,nelem,dtype=np.int32),
        function_space, mesh, material, fem_solver, Eulerx, Eulerp, processes= int(multiprocessing.cpu_count()/2))

    for elem in range(nelem):
        # COMPUATE LOCAL ELEMENTAL MASS MATRIX
        I_mass_elem, J_mass_elem, V_mass_elem = formulation.GetMassMatrix(elem,
            function_space, mesh, material, fem_solver, Eulerx, Eulerp)
        # SPARSE ASSEMBLY - MASS MATRIX
        SparseAssemblyNative(I_mass_elem,J_mass_elem,V_mass_elem,I_mass,J_mass,V_mass,
            elem,nvar,nodeperelem,mesh.elements)

    mass = csr_matrix((V_mass,(I_mass,J_mass)),shape=((nvar*mesh.points.shape[0],
        nvar*mesh.points.shape[0])),dtype=np.float64)

    return mass




#----------------- ASSEMBLY ROUTINE FOR TRACTION FORCES ONLY - FOR MODIFIED NEWTON RAPHSON ----------------------#
#----------------------------------------------------------------------------------------------------------------#


def AssembleInternalTractionForces(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp):

    # GET MESH DETAILS
    C = mesh.InferPolynomialDegree() - 1
    nvar = formulation.nvar
    ndim = formulation.ndim
    nelem = mesh.nelem
    nodeperelem = mesh.elements.shape[1]

    T = np.zeros((mesh.points.shape[0]*nvar,1),np.float64)
    # T = np.zeros((mesh.points.shape[0]*nvar,1),np.float32)

    for elem in range(nelem):

        t = formulation.GetElementalMatrices(elem,
            function_space, mesh, material, fem_solver, Eulerx, Eulerp)[3]

        # INTERNAL TRACTION FORCE ASSEMBLY
        # for iterator in range(0,nvar):
            # T[mesh.elements[elem,:]*nvar+iterator,0]+=t[iterator::nvar,0]
        RHSAssemblyNative(T,t,elem,nvar,nodeperelem,mesh.elements)


    return T



#------------------------------- ASSEMBLY ROUTINE FOR EXTERNAL TRACTION FORCES ----------------------------------#
#----------------------------------------------------------------------------------------------------------------#


def AssembleForces(boundary_condition, mesh, material, function_spaces, compute_traction_forces=True, compute_body_forces=False):

    Ft = np.zeros((mesh.points.shape[0]*material.nvar,1))
    Fb = np.zeros((mesh.points.shape[0]*material.nvar,1))

    if compute_traction_forces:
        Ft = AssembleExternalTractionForces(boundary_condition, mesh, material, function_spaces[-1])
    if compute_body_forces:
        Fb = AssembleBodyForces(boundary_condition, mesh, material, function_spaces[0])

    return Ft + Fb



def AssembleExternalTractionForces(boundary_condition, mesh, material, function_space):


    nvar = material.nvar
    ndim = material.ndim
    ngauss = function_space.AllGauss.shape[0]

    if ndim == 2:
        faces = mesh.edges
        nodeperelem = mesh.edges.shape[1]
    else:
        faces = mesh.faces
        nodeperelem = mesh.faces.shape[1]

    if boundary_condition.is_applied_neumann_shape_functions_computed is False:
        N = np.zeros((nodeperelem*nvar,nvar,ngauss))
        for i in range(nvar):
            N[i::nvar,i,:] = function_space.Bases
        boundary_condition.__Nt__ = N
        boundary_condition.is_applied_neumann_shape_functions_computed = True
    else:
        N = boundary_condition.__Nt__


    F = np.zeros((mesh.points.shape[0]*nvar,1))
    for face in range(faces.shape[0]):
        if boundary_condition.neumann_flags[face] == True:
            ElemTraction = boundary_condition.applied_neumann[face,:]
            external_traction = np.einsum("ijk,j,k->ik",N,ElemTraction,function_space.AllGauss[:,0]).sum(axis=1)
            RHSAssemblyNative(F,np.ascontiguousarray(external_traction[:,None]),face,nvar,nodeperelem,faces)


    # nvar = material.nvar
    # ndim = material.ndim

    # if ndim == 2:
    #     faces = np.copy(mesh.edges)
    #     nodeperelem = mesh.edges.shape[1]
    # else:
    #     faces = np.copy(mesh.faces)
    #     nodeperelem = mesh.faces.shape[1]

    # F = np.zeros((mesh.points.shape[0]*nvar,1))
    # for face in range(faces.shape[0]):
    #     if boundary_condition.neumann_flags[face] == True:
    #         ElemTraction = boundary_condition.applied_neumann[face,:]
    #         # LagrangeFaceCoords = mesh.points[faces[face,:],:]
    #         # ParentGradientX = np.einsum('ijk,jl->kil', function_space.Jm, LagrangeFaceCoords)
    #         # detJ = np.einsum('i,i->i',function_space.AllGauss[:,0],np.abs(np.linalg.det(ParentGradientX)))

    #         external_traction = np.zeros((nodeperelem*nvar))
    #         N = np.zeros((nodeperelem*nvar,nvar))
    #         for counter in range(function_space.AllGauss.shape[0]):
    #             for i in range(nvar):
    #                 N[i::nvar,i] = function_space.Bases[:,counter]

    #             external_traction += np.dot(N,ElemTraction)*function_space.AllGauss[counter,0]

    #        # RHS ASSEMBLY
    #         # for iterator in range(0,nvar):
    #             # F[faces[face,:]*nvar+iterator,0]+=external_traction[iterator::nvar]
    #         RHSAssemblyNative(F,np.ascontiguousarray(external_traction[:,None]),face,nvar,nodeperelem,faces)

    return F


def AssembleBodyForces(boundary_condition, mesh, material, function_space):


    nvar = material.nvar
    ndim = material.ndim
    nodeperelem = mesh.elements.shape[1]
    ngauss = function_space.AllGauss.shape[0]

    if boundary_condition.is_body_force_shape_functions_computed is False:
        N = np.zeros((nodeperelem*nvar,nvar,ngauss))
        for i in range(nvar):
            N[i::nvar,i,:] = function_space.Bases
        boundary_condition.__Nb__ = N
        boundary_condition.is_body_force_shape_functions_computed = True
    else:
        N = boundary_condition.__Nb__

    F = np.zeros((mesh.points.shape[0]*nvar,1))
    # BODY FORCE IS APPLIED IN THE Z-DIRECTION
    ElemTraction = np.zeros(nvar); ElemTraction[ndim-1] = -material.rho
    for elem in range(mesh.nelem):
        body_force = np.einsum("ijk,j,k->ik",N,ElemTraction,function_space.AllGauss[:,0]).sum(axis=1)
        RHSAssemblyNative(F,np.ascontiguousarray(body_force[:,None]),elem,nvar,nodeperelem,mesh.elements)


    # nvar = material.nvar
    # ndim = material.ndim
    # nodeperelem = mesh.elements.shape[1]

    # F = np.zeros((mesh.points.shape[0]*nvar,1))
    # # BODY FORCE IS APPLIED IN THE Z-DIRECTION
    # ElemTraction = np.zeros(nvar); ElemTraction[ndim-1] = -material.rho
    # for elem in range(mesh.nelem):
    #     # LagrangeElemCoords = mesh.points[mesh.elements[elem,:],:]

    #     body_force = np.zeros((nodeperelem*nvar))
    #     N = np.zeros((nodeperelem*nvar,nvar))
    #     for counter in range(function_space.AllGauss.shape[0]):
    #         for i in range(nvar):
    #             N[i::nvar,i] = function_space.Bases[:,counter]

    #         body_force += np.dot(N,ElemTraction)*function_space.AllGauss[counter,0]

    #    # RHS ASSEMBLY
    #     # for iterator in range(0,nvar):
    #         # F[faces[elem,:]*nvar+iterator,0]+=body_force[iterator::nvar]
    #     RHSAssemblyNative(F,np.ascontiguousarray(body_force[:,None]),elem,nvar,nodeperelem,mesh.elements)


    return F






#---------------------------------------- EXPLICIT ASSEMBLY ROUTINES --------------------------------------------#
#----------------------------------------------------------------------------------------------------------------#


def AssembleExplicit_NoLLD(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp):

    # GET MESH DETAILS
    C = mesh.InferPolynomialDegree() - 1
    nvar = formulation.nvar
    ndim = formulation.ndim
    nelem = mesh.nelem
    nodeperelem = mesh.elements.shape[1]

    T = np.zeros((mesh.points.shape[0]*nvar,1),np.float64)
    M = np.zeros((mesh.points.shape[0]*nvar,1),np.float64)

    mass, F = [], []
    if fem_solver.has_moving_boundary:
        F = np.zeros((mesh.points.shape[0]*nvar,1),np.float64)


    for elem in range(nelem):

        t, f, mass = formulation.GetElementalMatricesInVectorForm(elem,
                function_space, mesh, material, fem_solver, Eulerx, Eulerp)

        if fem_solver.has_moving_boundary:
            # RHS ASSEMBLY
            RHSAssemblyNative(F,f,elem,nvar,nodeperelem,mesh.elements)

        # LUMPED MASS ASSEMBLY
        if fem_solver.analysis_type != 'static' and fem_solver.is_mass_computed==False:
            RHSAssemblyNative(M,mass,elem,nvar,nodeperelem,mesh.elements)

        # INTERNAL TRACTION FORCE ASSEMBLY
        RHSAssemblyNative(T,t,elem,nvar,nodeperelem,mesh.elements)

    if fem_solver.analysis_type != 'static' and fem_solver.is_mass_computed==False:
        fem_solver.is_mass_computed = True

    return T, F, M



def AssembleExplicit(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp):

    if fem_solver.has_low_level_dispatcher and fem_solver.is_mass_computed is True:
        if not material.has_low_level_dispatcher:
            raise RuntimeError("Cannot dispatch to low level module, since material {} does not support it".format(type(material).__name__))

        if fem_solver.parallel:
            T = ExplicitParallelLauncher(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp)
            return T[:,None], [], []
        else:
            T, F, M = _LowLevelAssemblyExplicit_(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp)
            return T[:,None], F, M

    else:

        if fem_solver.has_low_level_dispatcher:
            if fem_solver.parallel:
                T = ExplicitParallelLauncher(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp)
            else:
                T = _LowLevelAssemblyExplicit_(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp)[0]
        else:
            return AssembleExplicit_NoLLD(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp)


        # GET MESH DETAILS
        nvar = formulation.nvar
        ndim = formulation.ndim
        nelem = mesh.nelem
        nodeperelem = mesh.elements.shape[1]

        F = []
        I_mass=[]; J_mass=[]; V_mass=[]
        if fem_solver.mass_type == "lumped":
            M = np.zeros((mesh.points.shape[0]*nvar,1),np.float64)
        else:
            # ALLOCATE VECTORS FOR SPARSE ASSEMBLY OF MASS MATRIX - CHANGE TYPES TO INT64 FOR DoF > 1e09
            I_mass=np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.int32)
            J_mass=np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.int32)
            V_mass=np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.float64)
            M = []

        for elem in range(nelem):

            LagrangeElemCoords = mesh.points[mesh.elements[elem,:],:]
            EulerElemCoords = Eulerx[mesh.elements[elem,:],:]
            if formulation.fields == "electro_mechanics":
                ElectricPotentialElem = Eulerp[mesh.elements[elem,:]]
            else:
                ElectricPotentialElem = []

            # COMPUTE THE MASS MATRIX
            if material.has_low_level_dispatcher:
                mass = formulation.__GetLocalMass_Efficient__(function_space,material,LagrangeElemCoords,EulerElemCoords,fem_solver,elem)
            else:
                mass = formulation.GetLocalMass_Efficient(function_space,material,LagrangeElemCoords,EulerElemCoords,fem_solver,elem)

            if fem_solver.mass_type == "lumped":
                mass = formulation.GetLumpedMass(mass)
                RHSAssemblyNative(M,mass,elem,nvar,nodeperelem,mesh.elements)
            else:
                # SPARSE ASSEMBLY - MASS MATRIX
                I_mass_elem, J_mass_elem, V_mass_elem = formulation.FindIndices(mass)
                SparseAssemblyNative(I_mass_elem,J_mass_elem,V_mass_elem,I_mass,J_mass,V_mass,
                    elem,nvar,nodeperelem,mesh.elements)

        # SET MASS FLAG HERE
        if fem_solver.is_mass_computed is False:
            if fem_solver.mass_type == "consistent":
                M = csr_matrix((V_mass,(I_mass,J_mass)),shape=((nvar*mesh.points.shape[0],
                nvar*mesh.points.shape[0])),dtype=np.float64)
            fem_solver.is_mass_computed = True


    return T[:,None], F, M





#---------------------------------------- PARALLEL ASSEMBLY ROUTINES --------------------------------------------#
#----------------------------------------------------------------------------------------------------------------#

class ImplicitParallelZipper(object):

    def __init__(self, fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp):
        self.fem_solver = fem_solver.__class__(analysis_type=fem_solver.analysis_type,
                                                analysis_nature=fem_solver.analysis_nature)
        self.function_space = function_space
        self.formulation = formulation
        self.mesh = mesh
        self.material = material
        self.Eulerx = Eulerx
        self.Eulerp = Eulerp

def ImplicitParallelExecuter_PoolBased(functor):
    return _LowLevelAssembly_Par_(functor.fem_solver, functor.function_space,
        functor.formulation, functor.mesh, functor.material, functor.Eulerx, functor.Eulerp)

def ImplicitParallelExecuter_ProcessBased(functor, proc, tups):
    tup = _LowLevelAssembly_Par_(functor.fem_solver, functor.function_space,
        functor.formulation, functor.mesh, functor.material, functor.Eulerx, functor.Eulerp)
    tups[proc] = tup
    # tups.append(tup) # FOR SERIAL CHECKS

def ImplicitParallelExecuter_ProcessQueueBased(functor, queue):
    tups = _LowLevelAssembly_Par_(functor.fem_solver, functor.function_space,
        functor.formulation, functor.mesh, functor.material, functor.Eulerx, functor.Eulerp)
    queue.put(tups)


def ImplicitParallelLauncher(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp):

    from multiprocessing import Process, Pool, Manager, Queue
    from contextlib import closing

    # GET MESH DETAILS
    nvar = formulation.nvar
    ndim = formulation.ndim
    nelem = mesh.nelem
    nnode = mesh.points.shape[0]
    nodeperelem = mesh.elements.shape[1]
    local_capacity = int((nvar*nodeperelem)**2)

    pmesh, pelement_indices, pnode_indices, partitioned_maps = fem_solver.pmesh, \
        fem_solver.pelement_indices, fem_solver.pnode_indices, fem_solver.partitioned_maps


    # ALLOCATE VECTORS FOR SPARSE ASSEMBLY OF STIFFNESS MATRIX - CHANGE TYPES TO INT64 FOR DoF > 1e09
    I_stiffness=np.zeros((nelem,local_capacity),dtype=np.int32)
    J_stiffness=np.zeros((nelem,local_capacity),dtype=np.int32)
    V_stiffness=np.zeros((nelem,local_capacity),dtype=np.float64)

    I_mass=[]; J_mass=[]; V_mass=[]
    if fem_solver.analysis_type !='static' and fem_solver.is_mass_computed is False:
        # ALLOCATE VECTORS FOR SPARSE ASSEMBLY OF MASS MATRIX - CHANGE TYPES TO INT64 FOR DoF > 1e09
        I_mass=np.zeros((nelem,local_capacity),dtype=np.int32)
        J_mass=np.zeros((nelem,local_capacity),dtype=np.int32)
        V_mass=np.zeros((nelem,local_capacity),dtype=np.float64)

    T = np.zeros((mesh.points.shape[0],nvar),np.float64)

    funcs = []
    for proc in range(fem_solver.no_of_cpu_cores):
        pnodes = pnode_indices[proc]
        Eulerx_current = Eulerx[pnodes,:]
        Eulerp_current = Eulerp[pnodes]
        funcs.append(ImplicitParallelZipper(fem_solver, function_space, formulation,
            pmesh[proc], material, Eulerx_current, Eulerp_current))

    # # SERIAL
    # tups = []
    # for i in range(fem_solver.no_of_cpu_cores):
    #     ImplicitParallelExecuter_ProcessBased(funcs[i], i, tups)
    # for i in range(fem_solver.no_of_cpu_cores):
    #     pnodes = pnode_indices[i]
    #     pelements = pelement_indices[i]
    #     I_stiffness[pelements,:] = partitioned_maps[i][tups[i][0]].reshape(pmesh[i].nelem,local_capacity)
    #     J_stiffness[pelements,:] = partitioned_maps[i][tups[i][1]].reshape(pmesh[i].nelem,local_capacity)
    #     V_stiffness[pelements,:] = tups[i][2].reshape(pmesh[i].nelem,local_capacity)
    #     T[pnodes,:] += tups[i][-1].reshape(pnodes.shape[0],nvar)

    #     if fem_solver.analysis_type != "static" and fem_solver.is_mass_computed is False:
    #         I_stiffness[pelements,:] = partitioned_maps[i][tups[i][3]].reshape(pmesh[i].nelem,local_capacity)
    #         J_stiffness[pelements,:] = partitioned_maps[i][tups[i][4]].reshape(pmesh[i].nelem,local_capacity)
    #         V_stiffness[pelements,:] = tups[i][5].reshape(pmesh[i].nelem,local_capacity)


    # POOL BASED
    if fem_solver.parallel_model == "pool":
        with closing(Pool(processes=fem_solver.no_of_cpu_cores)) as pool:
            tups = pool.map(ImplicitParallelExecuter_PoolBased,funcs)
            pool.terminate()

    # JOBLIB BASED
    elif fem_solver.parallel_model == "joblib":
        try:
            from joblib import Parallel, delayed
        except ImportError:
            raise ImportError("Joblib is not installed. Install it 'using pip install joblib'")
        tups = Parallel(n_jobs=fem_solver.no_of_cpu_cores)(delayed(ImplicitParallelExecuter_PoolBased)(func) for func in funcs)
        # tups = Parallel(n_jobs=10, backend="threading")(delayed(ImplicitParallelExecuter_PoolBased)(func) for func in funcs)

    # SCOOP BASED
    elif fem_solver.parallel_model == "scoop":
        try:
            from scoop import futures
        except ImportError:
            raise ImportError("Scoop is not installed. Install it 'using pip install scoop'")
        # tups = futures.map(ImplicitParallelExecuter_PoolBased, funcs)
        tups = list(futures.map(ImplicitParallelExecuter_PoolBased, funcs))

    # PROCESS AND MANAGER BASED
    elif fem_solver.parallel_model == "context_manager":
        procs = []
        manager = Manager(); tups = manager.dict() # SPAWNS A NEW PROCESS
        for i, func in enumerate(funcs):
            proc = Process(target=ImplicitParallelExecuter_ProcessBased, args=(func,i,tups))
            procs.append(proc)
            proc.start()
        for proc in procs:
            proc.join()

    # PROCESS AND QUEUE BASED
    elif fem_solver.parallel_model == "queue":
        procs = []
        for i, func in enumerate(funcs):
            queue = Queue()
            proc = Process(target=ImplicitParallelExecuter_ProcessQueueBased, args=(func,queue))
            proc.daemon = True
            procs.append(proc)
            proc.start()
            tups = queue.get()
            pnodes = pnode_indices[i]
            pelements = pelement_indices[i]
            I_stiffness[pelements,:] = partitioned_maps[i][tups[0]].reshape(pmesh[i].nelem,local_capacity)
            J_stiffness[pelements,:] = partitioned_maps[i][tups[1]].reshape(pmesh[i].nelem,local_capacity)
            V_stiffness[pelements,:] = tups[2].reshape(pmesh[i].nelem,local_capacity)
            T[pnodes,:] += tups[-1].reshape(pnodes.shape[0],nvar)

            if fem_solver.analysis_type != "static" and fem_solver.is_mass_computed is False:
                I_mass[pelements,:] = partitioned_maps[i][tups[3]].reshape(pmesh[i].nelem,local_capacity)
                J_mass[pelements,:] = partitioned_maps[i][tups[4]].reshape(pmesh[i].nelem,local_capacity)
                V_mass[pelements,:] = tups[5].reshape(pmesh[i].nelem,local_capacity)
            proc.join()

    if fem_solver.parallel_model == "pool" or fem_solver.parallel_model == "context_manager" \
        or fem_solver.parallel_model == "joblib" or fem_solver.parallel_model == "scoop":
        for i in range(fem_solver.no_of_cpu_cores):
            pnodes = pnode_indices[i]
            pelements = pelement_indices[i]
            I_stiffness[pelements,:] = partitioned_maps[i][tups[i][0]].reshape(pmesh[i].nelem,local_capacity)
            J_stiffness[pelements,:] = partitioned_maps[i][tups[i][1]].reshape(pmesh[i].nelem,local_capacity)
            V_stiffness[pelements,:] = tups[i][2].reshape(pmesh[i].nelem,local_capacity)
            T[pnodes,:] += tups[i][-1].reshape(pnodes.shape[0],nvar)

            if fem_solver.analysis_type != "static" and fem_solver.is_mass_computed is False:
                I_mass[pelements,:] = partitioned_maps[i][tups[i][3]].reshape(pmesh[i].nelem,local_capacity)
                J_mass[pelements,:] = partitioned_maps[i][tups[i][4]].reshape(pmesh[i].nelem,local_capacity)
                V_mass[pelements,:] = tups[i][5].reshape(pmesh[i].nelem,local_capacity)


    stiffness = csr_matrix((V_stiffness.ravel(),(I_stiffness.ravel(),J_stiffness.ravel())),
        shape=((nvar*mesh.points.shape[0],nvar*mesh.points.shape[0])),dtype=np.float64)

    F, mass = [], []

    if fem_solver.analysis_type != "static" and fem_solver.is_mass_computed is False:
        mass = csr_matrix((V_mass.ravel(),(I_mass.ravel(),J_mass.ravel())),
            shape=((nvar*mesh.points.shape[0],nvar*mesh.points.shape[0])),dtype=np.float64)


    return stiffness, T.ravel(), F, mass

















class ExplicitParallelZipper(object):
    def __init__(self, function_space, formulation, mesh, material, Eulerx, Eulerp):
        self.function_space = function_space
        self.formulation = formulation
        self.mesh = mesh
        self.material = material
        self.Eulerx = Eulerx
        self.Eulerp = Eulerp

class ExplicitParallelZipperMPI(object):
    def __init__(self, formulation, mesh, material, pnodes):
        self.formulation = formulation
        self.material = material
        self.mesh = mesh
        self.pnodes = pnodes

class ExplicitParallelZipperHDF5(object):
    def __init__(self, formulation, mesh, material):
        self.formulation = formulation
        self.material = material
        self.mesh = mesh


def ExplicitParallelExecuter_PoolBased(functor):
    return _LowLevelAssemblyExplicit_Par_(functor.function_space,
        functor.formulation, functor.mesh, functor.material, functor.Eulerx, functor.Eulerp)

def ExplicitParallelExecuter_ProcessBased(functor, proc, Ts):
    T = _LowLevelAssemblyExplicit_Par_(functor.function_space,
        functor.formulation, functor.mesh, functor.material, functor.Eulerx, functor.Eulerp)
    Ts[proc] = T

def ExplicitParallelExecuter_ProcessQueueBased(functor, queue):
    T = _LowLevelAssemblyExplicit_Par_(functor.function_space,
        functor.formulation, functor.mesh, functor.material, functor.Eulerx, functor.Eulerp)
    queue.put(T)

def ExplicitParallelExecuter_HDF5Based(functor, proc, fname_in, fname_out):

    import h5py

    h5f_out = h5py.File(fname_out+str(proc)+'.h5','r')
    Eulerx = h5f_out['Geometry']['Eulerx'][:]
    Eulerp = h5f_out['Geometry']['Eulerp'][:]
    functor.mesh.points = h5f_out['Geometry']['points'][:]
    functor.mesh.elements = h5f_out['Geometry']['elements'][:]

    T = _LowLevelAssemblyExplicit_Par_(functor.formulation.function_spaces[0],
        functor.formulation, functor.mesh, functor.material, Eulerx, Eulerp)

    # T = _LowLevelAssemblyExplicit_Par_(functor.formulation.function_spaces[0],
    #     functor.formulation, functor.mesh, functor.material, functor.Eulerx, functor.Eulerp)

    h5f = h5py.File(fname_in+str(proc)+'.h5','w')
    h5f.create_dataset('T', data=T)
    h5f.close()


def ExplicitParallelLauncher(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp):

    from multiprocessing import Process, Pool, Manager, Queue
    from contextlib import closing

    pmesh, pelement_indices, pnode_indices = fem_solver.pmesh, fem_solver.pelement_indices, fem_solver.pnode_indices
    T_all = np.zeros((mesh.points.shape[0],formulation.nvar),np.float64)

    # MPI BASED
    if fem_solver.parallel_model == "mpi":
        try:
            from mpi4py import MPI
        except ImportError:
            raise ImportError("mpi4py is not installed. Install it using 'pip install mpi4py'")
        from Florence import PWD
        comm = MPI.COMM_SELF.Spawn(sys.executable,
                                   args=[PWD(__file__)+'/MPIParallelExplicitAssembler.py'],
                                   maxprocs=fem_solver.no_of_cpu_cores)

        funcs = []
        for proc in range(fem_solver.no_of_cpu_cores):
            obj = ExplicitParallelZipperMPI(formulation, pmesh[proc], material, pnode_indices[proc])
            funcs.append(obj)

        T_all_size = np.array([mesh.points.shape[0],formulation.ndim, formulation.nvar],dtype="i")
        comm.Bcast([T_all_size, MPI.INT], root=MPI.ROOT)
        comm.bcast(funcs, root=MPI.ROOT)
        comm.Bcast([Eulerx, MPI.DOUBLE], root=MPI.ROOT)
        comm.Bcast([Eulerp, MPI.DOUBLE], root=MPI.ROOT)

        # for proc in range(fem_solver.no_of_cpu_cores):
        #     globals()['points%s' % proc] = pmesh[proc].points
        #     globals()['elements%s' % proc] = pmesh[proc].elements
        #     globals()['nelems%s' % proc] = pmesh[proc].elements.nelem
        #     globals()['nnodes%s' % proc] = pmesh[proc].points.nnode

        # Main T_all TO BE FILLED
        comm.Reduce(None, [T_all, MPI.DOUBLE], root=MPI.ROOT)

        comm.Disconnect()

        return T_all.ravel()


    # PROCESS AND HDF5 BASED
    elif fem_solver.parallel_model == "hdf5":

        try:
            import h5py
        except ImportError:
            raise ImportError("h5py is not installed. Install it using 'pip install h5py'")
        import shutil
        from Florence import Mesh

        home = os.path.expanduser("~")
        tmp_folder = os.path.join(home,".florence_tmp000")
        if not os.path.exists(tmp_folder):
            os.makedirs(tmp_folder)

        fname_in = os.path.join(tmp_folder,"results_explicit")
        fname_out = os.path.join(tmp_folder,"geometry_explicit")

        # funcs = []
        # for proc in range(fem_solver.no_of_cpu_cores):
        #     pnodes = pnode_indices[proc]
        #     Eulerx_current = Eulerx[pnodes,:]
        #     Eulerp_current = Eulerp[pnodes]
        #     obj = ExplicitParallelZipper(function_space, formulation,
        #         pmesh[proc], material, Eulerx_current, Eulerp_current)
        #     funcs.append(obj)

        funcs = []
        for proc in range(fem_solver.no_of_cpu_cores):
            pnodes = pnode_indices[proc]
            Eulerx_current = Eulerx[pnodes,:]
            Eulerp_current = Eulerp[pnodes]


            h5f_out = h5py.File(fname_out+str(proc)+'.h5','w')
            grp = h5f_out.create_group('Geometry')

            grp.create_dataset('elements', data=pmesh[proc].elements)
            grp.create_dataset('points', data=pmesh[proc].points)
            grp.create_dataset('Eulerx', data=Eulerx_current)
            grp.create_dataset('Eulerp', data=Eulerp_current)

            h5f_out.close()

            imesh = Mesh()
            imesh.nnode, imesh.nelem, imesh.element_type = pmesh[proc].nnode, pmesh[proc].nelem, pmesh[proc].element_type

            obj = ExplicitParallelZipperHDF5(formulation, imesh, material)
            funcs.append(obj)

        procs = []
        for i, func in enumerate(funcs):
            proc = Process(target=ExplicitParallelExecuter_HDF5Based, args=(func, i, fname_in, fname_out))
            procs.append(proc)
            proc.start()
        for proc in procs:
            proc.join()

        for proc in range(fem_solver.no_of_cpu_cores):
            h5f = h5py.File(fname_in+str(proc)+'.h5','r')
            T = h5f['T'][:]
            pnodes = pnode_indices[proc]
            T_all[pnodes,:] += T.reshape(pnodes.shape[0],formulation.nvar)

        shutil.rmtree(tmp_folder)

        return T_all.ravel()




    funcs = []
    for proc in range(fem_solver.no_of_cpu_cores):
        pnodes = pnode_indices[proc]
        Eulerx_current = Eulerx[pnodes,:]
        Eulerp_current = Eulerp[pnodes]
        obj = ExplicitParallelZipper(function_space, formulation,
            pmesh[proc], material, Eulerx_current, Eulerp_current)
        funcs.append(obj)

    # SERIAL
    # for proc in range(fem_solver.no_of_cpu_cores):
    #     T = ExplicitParallelExecuter_PoolBased(funcs[proc])
    #     pnodes = pnode_indices[proc]
    #     T_all[pnodes,:] += T.reshape(pnodes.shape[0],formulation.nvar)


    # PROCESS AND MANAGER BASED
    if fem_solver.parallel_model == "context_manager":
        procs = []
        manager = Manager(); Ts = manager.dict() # SPAWNS A NEW PROCESS
        for i, func in enumerate(funcs):
            proc = Process(target=ExplicitParallelExecuter_ProcessBased, args=(func,i,Ts))
            procs.append(proc)
            proc.start()
        for proc in procs:
            proc.join()

    # POOL BASED
    elif fem_solver.parallel_model == "pool":
        with closing(Pool(processes=fem_solver.no_of_cpu_cores)) as pool:
            Ts = pool.map(ExplicitParallelExecuter_PoolBased,funcs)
            pool.terminate()
            # DOESN'T SCALE WELL
            # Ts = pool.map_async(ExplicitParallelExecuter_PoolBased,funcs)
            # Ts.wait()
            # Ts = Ts.get()

    # JOBLIB BASED
    elif fem_solver.parallel_model == "joblib":
        try:
            from joblib import Parallel, delayed
        except ImportError:
            raise ImportError("Joblib is not installed. Install it using 'pip install joblib'")
        Ts = Parallel(n_jobs=fem_solver.no_of_cpu_cores)(delayed(ExplicitParallelExecuter_PoolBased)(func) for func in funcs)
        # Ts = Parallel(n_jobs=10, backend="threading")(delayed(ImplicitParallelExecuter_PoolBased)(func) for func in funcs)

    # SCOOP BASED
    elif fem_solver.parallel_model == "scoop":
        try:
            from scoop import futures
        except ImportError:
            raise ImportError("Scoop is not installed. Install it using 'pip install scoop'")
        Ts = list(futures.map(ExplicitParallelExecuter_PoolBased, funcs))

    # PROCESS AND QUEUE BASED
    elif fem_solver.parallel_model == "queue":
        procs = []
        for i, func in enumerate(funcs):
            queue = Queue()
            proc = Process(target=ExplicitParallelExecuter_ProcessQueueBased, args=(func,queue))
            proc.daemon = True
            procs.append(proc)
            proc.start()
            pnodes = pnode_indices[i]
            T = queue.get()
            T_all[pnodes,:] += T.reshape(pnodes.shape[0],formulation.nvar)
            proc.join()



    if fem_solver.parallel_model == "pool" or fem_solver.parallel_model == "context_manager" \
        or fem_solver.parallel_model == "joblib" or fem_solver.parallel_model == "scoop":

        for proc in range(fem_solver.no_of_cpu_cores):
            pnodes = pnode_indices[proc]
            T_all[pnodes,:] += Ts[proc].reshape(pnodes.shape[0],formulation.nvar)



    return T_all.ravel()



from __future__ import print_function
import numpy as np, scipy as sp, sys, os, gc
from warnings import warn
from time import time

from Florence.QuadratureRules import GaussLobattoQuadrature
from Florence.QuadratureRules.FeketePointsTri import FeketePointsTri
from Florence.QuadratureRules.EquallySpacedPoints import EquallySpacedPoints
from Florence.QuadratureRules import GaussLobattoPointsQuad

class BoundaryCondition(object):
    """Base class for applying all types of boundary conditions"""

    def __init__(self,
        surface_identification_algorithm='minimisation',
        modify_linear_mesh_on_projection=False,
        project_on_curves=True,
        activate_bounding_box=False,
        bounding_box_padding=1e-3,
        has_planar_surfaces=True,
        solve_for_planar_faces=True,
        save_dirichlet_data=False,
        save_nurbs_data=False,
        filename=None,
        read_dirichlet_from_file=False,
        make_loading="ramp"
        ):

        # TYPE OF BOUNDARY: straight or nurbs
        self.boundary_type = 'straight'
        self.dirichlet_data_applied_at = 'node' # or 'faces'
        self.neumann_data_applied_at = 'node' # or 'faces'
        self.requires_cad = False
        self.cad_file = None
        # PROJECTION TYPE FOR CAD EITHER orthogonal OR arc_length
        self.projection_type = 'orthogonal'
        # WHAT TYPE OF ARC LENGTH BASED PROJECTION, EITHER 'equal' OR 'fekete'
        self.nodal_spacing_for_cad = 'equal'
        self.project_on_curves = project_on_curves
        self.scale_mesh_on_projection = False
        self.scale_value_on_projection = 1.0
        self.condition_for_projection = 1.0e20
        self.has_planar_surfaces = False
        self.solve_for_planar_faces = solve_for_planar_faces
        self.projection_flags = None
        # FIX DEGREES OF FREEDOM EVERY WHERE CAD PROJECTION IS NOT APPLIED
        self.fix_dof_elsewhere = True
        # FOR 3D ARC-LENGTH PROJECTION
        self.orthogonal_fallback_tolerance = 1.0
        # WHICH ALGORITHM TO USE FOR SURFACE IDENTIFICATION, EITHER 'minimisation' or 'pure_projection'
        self.surface_identification_algorithm = surface_identification_algorithm
        # MODIFY LINEAR MESH ON PROJECTION
        self.modify_linear_mesh_on_projection = modify_linear_mesh_on_projection
        # COMPUTE A BOUNDING BOX FOR EACH CAD SURFACE
        self.activate_bounding_box = activate_bounding_box
        self.bounding_box_padding = float(bounding_box_padding)

        # FOR IGAKit WRAPPER
        self.nurbs_info = None
        self.nurbs_condition = None

        self.analysis_type = 'static'
        self.analysis_nature = 'linear'

        self.is_dirichlet_computed = False
        self.columns_out = None
        self.columns_in = None
        self.applied_dirichlet = None
        self.save_dirichlet_data = save_dirichlet_data
        self.save_nurbs_data = save_nurbs_data
        self.filename = filename
        self.read_dirichlet_from_file = read_dirichlet_from_file

        self.dirichlet_flags = None
        self.neumann_flags = None
        self.applied_neumann = None
        self.is_applied_neumann_shape_functions_computed = False
        self.is_body_force_shape_functions_computed = False

        self.make_loading = make_loading # "ramp" or "constant"


        # NODAL FORCES GENERATED BASED ON DIRICHLET OR NEUMANN ARE NOT
        # IMPLEMENTED AS PART OF BOUNDARY CONDITION YET. THIS ESSENTIALLY
        # MEANS SAVING MULTIPLE RHS VALUES
        # self.dirichlet_forces = None
        # self.neumann_forces = None

        # # THE FOLLOWING MEMBERS ARE NOT UPDATED, TO REDUCE MEMORY FOOTPRINT
        # self.external_nodal_forces = None
        # self.internal_traction_forces = None
        # self.residual = None


    def SetAnalysisParameters(self,analysis_type='static',analysis_nature='linear',
        columns_in=None,columns_out=None,applied_dirichlet=None):
        """Set analysis parameters such as analysis type, analysis nature and even
            Dirichlet boundary conditions if known a priori
        """
        self.analysis_type = analysis_type
        self.analysis_nature = analysis_nature
        self.columns_out = columns_out
        self.columns_in = columns_in
        self.applied_dirichlet = applied_dirichlet



    def SetCADProjectionParameters(self, cad_file=None, requires_cad=True, projection_type='orthogonal',
        nodal_spacing='equal', project_on_curves=True, has_planar_surfaces=True, solve_for_planar_faces=True,
        scale=1.0,condition=1.0e20, projection_flags=None, fix_dof_elsewhere=True,
        orthogonal_fallback_tolerance=1.0, surface_identification_algorithm='minimisation',
        modify_linear_mesh_on_projection=False, activate_bounding_box=False, bounding_box_padding=1e-3):
        """Set parameters for CAD projection in order to obtain dirichlet boundary
            conditinos
        """

        self.boundary_type = 'nurbs'
        self.requires_cad = requires_cad
        self.cad_file = cad_file
        self.projection_type = projection_type
        self.scale_mesh_on_projection = True
        self.scale_value_on_projection = 1.0*scale
        self.condition_for_projection = 1.0*condition
        self.project_on_curves = project_on_curves
        self.has_planar_surfaces = has_planar_surfaces
        self.solve_for_planar_faces = solve_for_planar_faces
        self.projection_flags = projection_flags
        self.fix_dof_elsewhere = fix_dof_elsewhere
        self.orthogonal_fallback_tolerance = orthogonal_fallback_tolerance
        self.surface_identification_algorithm = surface_identification_algorithm
        self.modify_linear_mesh_on_projection = modify_linear_mesh_on_projection
        self.nodal_spacing_for_cad = nodal_spacing
        self.activate_bounding_box = activate_bounding_box
        self.bounding_box_padding = float(bounding_box_padding)

        self.project_on_curves = int(self.project_on_curves)
        self.modify_linear_mesh_on_projection = int(self.modify_linear_mesh_on_projection)


    def SetProjectionCriteria(self, proj_func, mesh, *args, **kwargs):
        """Factory function for setting projection criteria specific
            to a problem

            input:
                func                [function] function that computes projection criteria
                mesh                [Mesh] an instance of mesh class
        """


        self.projection_flags = proj_func(mesh, *args, **kwargs)

        if isinstance(self.projection_flags,np.ndarray):
            if self.projection_flags.ndim==1:
                self.projection_flags.reshape(-1,1)
                ndim = mesh.InferSpatialDimension()
                if self.projection_flags.shape[0] != mesh.edges.shape[0] and ndim == 2:
                    raise ValueError("Projection flags are incorrect. "
                        "Ensure that your projection function returns an ndarray of shape (mesh.edges.shape[0],1)")
                elif self.projection_flags.shape[0] != mesh.faces.shape[0] and ndim == 3:
                    raise ValueError("Projection flags are incorrect. "
                        "Ensure that your projection function returns an ndarray of shape (mesh.faces.shape[0],1)")
        else:
            raise ValueError("Projection flags for CAD not set. "
                "Ensure that your projection function returns an ndarray")


    def GetProjectionCriteria(self,mesh):
        """Convenience method for computing projection flags, as many problems
        require this type of projection
        """

        ndim = mesh.InferSpatialDimension()
        if ndim==3:
            boundaries = mesh.faces
        elif ndim==2:
            boundaries = mesh.edges

        projection_flags = np.zeros((boundaries.shape[0],1),dtype=np.uint64)
        num = boundaries.shape[1]

        xyz = (self.scale_value_on_projection/num)*np.sum(mesh.points[boundaries,:],axis=1)
        projection_flags[:,0] = np.linalg.norm(xyz,axis=1) < self.condition_for_projection
        self.projection_flags = projection_flags


    def GetGeometryMeshScale(self,gpoints,mesh):
        """Compares CAD geometry and mesh, to check if the mesh coordinates
            require scaling

            raises an error if scaling is
        """

        gminx, gmaxx = np.min(gpoints[:,0]), np.max(gpoints[:,0])
        gminy, gmaxy = np.min(gpoints[:,1]), np.max(gpoints[:,1])
        gmax = np.max(gpoints)
        mmax = np.max(mesh.points)

        # NOTE THAT THE BOUNDS OF NURBS BOUNDARY ISN'T
        # NECESSARILY EXACTLY THE SAME AS THE MESH, EVEN IF
        # SCALED APPROPRIATELY
        gbounds = np.array([[gminx,gminy],[gmaxx,gmaxy]])

        units_scalar = [1000.,25.4]
        for scale in units_scalar:
            if np.isclose(gmax/mmax,scale):
                if self.scale_value_on_projection != scale:
                    self.scale_value_on_projection = scale
                    raise ValueError('Geometry to mesh scale seems incorrect. Change it to %9.3f' % scale)
                    # A SIMPLE WARNING IS NOT POSSIBLE AT THE MOMENT
                    # warn('Geometry to mesh scale seems incorrect. Change it to %9.3f' % scale)
                    # break

        return gbounds



    def SetDirichletCriteria(self, func, *args, **kwargs):
        self.dirichlet_flags = func(*args, **kwargs)
        return self.dirichlet_flags


    def SetNeumannCriteria(self, func, *args, **kwargs):
        tups = func(*args, **kwargs)
        if not isinstance(tups,tuple) and self.neumann_data_applied_at == "node":
            self.neumann_flags = tups
            return self.neumann_flags
        else:
            self.neumann_data_applied_at == "face"
            tups = func(*args, **kwargs)
            if len(tups) !=2:
                raise ValueError("User-defined Neumann criterion function {} should return one flag and one data array".format(func.__name__))
            self.neumann_flags = tups[0]
            self.applied_neumann = tups[1]
            return tups


    def GetDirichletBoundaryConditions(self, formulation, mesh, material=None, solver=None, fem_solver=None):

        nvar = formulation.nvar
        ndim = formulation.ndim
        self.columns_in, self.applied_dirichlet = [], []


        #----------------------------------------------------------------------------------------------------#
        #-------------------------------------- NURBS BASED SOLUTION ----------------------------------------#
        #----------------------------------------------------------------------------------------------------#
        if self.boundary_type == 'nurbs':

            tCAD = time()

            if self.read_dirichlet_from_file is False:

                if not self.is_dirichlet_computed:
                    # GET DIRICHLET BOUNDARY CONDITIONS BASED ON THE EXACT GEOMETRY FROM CAD
                    if self.requires_cad:
                        # CALL POSTMESH WRAPPER
                        nodesDBC, Dirichlet = self.PostMeshWrapper(formulation, mesh, material, solver, fem_solver)
                else:
                    nodesDBC, Dirichlet = self.nodesDBC, self.Dirichlet


                # GET DIRICHLET DoFs
                self.columns_out = (np.repeat(nodesDBC,nvar,axis=1)*nvar +\
                 np.tile(np.arange(nvar)[None,:],nodesDBC.shape[0]).reshape(nodesDBC.shape[0],formulation.ndim)).ravel()
                self.applied_dirichlet = Dirichlet.ravel()


                # FIX THE DOF IN THE REST OF THE BOUNDARY
                if self.fix_dof_elsewhere:
                    if ndim==2:
                        rest_dofs = np.setdiff1d(np.unique(mesh.edges),nodesDBC)
                    elif ndim==3:
                        rest_dofs = np.setdiff1d(np.unique(mesh.faces),nodesDBC)

                    rest_out = np.repeat(rest_dofs,nvar)*nvar + np.tile(np.arange(nvar),rest_dofs.shape[0])
                    rest_app = np.zeros(rest_dofs.shape[0]*nvar)

                    self.columns_out = np.concatenate((self.columns_out,rest_out)).astype(np.int64)
                    self.applied_dirichlet = np.concatenate((self.applied_dirichlet,rest_app))


                print('Finished identifying Dirichlet boundary conditions from CAD geometry.',
                    ' Time taken', time()-tCAD, 'seconds')

            else:

                end = -3
                self.applied_dirichlet = np.loadtxt(mesh.filename.split(".")[0][:end]+"_Dirichlet_"+"P"+str(MainData.C+1)+".dat",dtype=np.float64)
                self.columns_out = np.loadtxt(mesh.filename.split(".")[0][:end]+"_ColumnsOut_"+"P"+str(MainData.C+1)+".dat")

                print('Finished identifying Dirichlet boundary conditions from CAD geometry.',
                    ' Time taken', time()-tCAD, 'seconds')

        #----------------------------------------------------------------------------------------------------#
        #------------------------------------- NON-NURBS BASED SOLUTION -------------------------------------#
        #----------------------------------------------------------------------------------------------------#

        elif self.boundary_type == 'straight' or self.boundary_type == 'mixed':
            # IF DIRICHLET BOUNDARY CONDITIONS ARE APPLIED DIRECTLY AT NODES
            if self.dirichlet_flags is None:
                raise RuntimeError("Dirichlet boundary conditions are not set for the analysis")

            if self.dirichlet_data_applied_at == 'node':
                if self.analysis_type == "dynamic":
                    # FOR DYNAMIC ANALYSIS IT IS ASSUMED THAT
                    # self.columns_in and self.columns_out DO NOT CHANGE
                    # DURING THE ANALYSIS
                    if self.dirichlet_flags.ndim == 3:
                        flat_dirich = self.dirichlet_flags[:,:,0].ravel()
                        self.columns_out = np.arange(self.dirichlet_flags[:,:,0].size)[~np.isnan(flat_dirich)]
                        self.applied_dirichlet = np.zeros((self.columns_out.shape[0],self.dirichlet_flags.shape[2]))

                        for step in range(self.dirichlet_flags.shape[2]):
                            flat_dirich = self.dirichlet_flags[:,:,step].ravel()
                            self.applied_dirichlet[:,step] = flat_dirich[~np.isnan(flat_dirich)]
                    elif self.dirichlet_flags.ndim == 2:
                        flat_dirich = self.dirichlet_flags.ravel()
                        self.columns_out = np.arange(self.dirichlet_flags.size)[~np.isnan(flat_dirich)]
                        self.applied_dirichlet = flat_dirich[~np.isnan(flat_dirich)]
                    else:
                        raise ValueError("Incorrect Dirichlet flags for dynamic analysis")

                else:
                    flat_dirich = self.dirichlet_flags.ravel()
                    self.columns_out = np.arange(self.dirichlet_flags.size)[~np.isnan(flat_dirich)]
                    self.applied_dirichlet = flat_dirich[~np.isnan(flat_dirich)]

        # GENERAL PROCEDURE - GET REDUCED MATRICES FOR FINAL SOLUTION
        self.columns_out = self.columns_out.astype(np.int64)
        self.columns_in = np.delete(np.arange(0,nvar*mesh.points.shape[0]),self.columns_out)

        if self.save_dirichlet_data:
            from scipy.io import savemat
            diri_dict = {'columns_in':self.columns_in,
                'columns_out':self.columns_out,
                'applied_dirichlet':self.applied_dirichlet}
            savemat(self.filename,diri_dict, do_compression=True)



    def ConvertStaticsToDynamics(self, mesh, nincr):
        """Convert static boundary condition data to dynamic
        """

        if self.analysis_type == "dynamic":
            # AVOID ZERO DIVISION FOR RAMP (LINSPACE TYPE) LOADING
            nincr_last = float(nincr-1) if nincr !=1 else 1
            if self.boundary_type != "nurbs":
                if self.dirichlet_flags is not None:
                    if self.dirichlet_flags.ndim == 2:
                        dum = np.zeros((self.dirichlet_flags.shape[0],self.dirichlet_flags.shape[1],nincr))
                        for incr in range(nincr):
                            if self.make_loading == "constant":
                                dum[:,:,incr] = self.dirichlet_flags/float(nincr)
                            else:
                                dum[:,:,incr] = incr*self.dirichlet_flags/nincr_last
                        self.dirichlet_flags = np.copy(dum)
                    else:
                        return
            else:
                if self.applied_dirichlet is not None:
                    if self.applied_dirichlet.ndim == 1:
                        dum = np.zeros((self.applied_dirichlet.shape[0],nincr))
                        for incr in range(nincr):
                            if self.make_loading == "constant":
                                dum[:,incr] = self.applied_dirichlet/float(nincr)
                            else:
                                dum[:,incr] = incr*self.applied_dirichlet/nincr_last
                        self.applied_dirichlet = np.copy(dum)
                    else:
                        return


            if self.neumann_flags is not None:

                ndim = mesh.InferSpatialDimension()
                if self.neumann_flags.shape[0] == mesh.points.shape[0]:
                    self.neumann_data_applied_at = "node"
                else:
                    if ndim==3:
                        if self.neumann_flags.shape[0] == mesh.faces.shape[0]:
                            self.neumann_data_applied_at = "face"
                    elif ndim==2:
                        if self.neumann_flags.shape[0] == mesh.edges.shape[0]:
                            self.neumann_data_applied_at = "face"

                if self.neumann_data_applied_at == "node":
                    if self.neumann_flags.ndim == 2:
                        dum = np.zeros((self.neumann_flags.shape[0],self.neumann_flags.shape[1],nincr))
                        for incr in range(nincr):
                            if self.make_loading == "constant":
                                dum[:,:,incr] = self.neumann_flags/float(nincr)
                            else:
                                dum[:,:,incr] = incr*self.neumann_flags/nincr_last
                        self.neumann_flags = np.copy(dum)
                    else:
                        return
                elif self.neumann_data_applied_at == "face":
                    if self.applied_neumann is None:
                        raise ValueError("Incorrect Neumann data supplied")
                    if self.neumann_flags.ndim == 1:
                        tmp_flags = np.zeros((self.neumann_flags.shape[0],nincr))
                        tmp_data = np.zeros((self.applied_neumann.shape[0],self.applied_neumann.shape[1],nincr))
                        for incr in range(nincr):
                            if self.make_loading == "constant":
                                tmp_data[:,:,incr] = self.applied_neumann/float(nincr)
                            else:
                                tmp_data[:,:,incr] = incr*self.applied_neumann/nincr_last
                            tmp_flags[:,incr] = self.neumann_flags
                        self.neumann_flags = np.copy(tmp_flags)
                        self.applied_neumann = np.copy(tmp_data)
                    else:
                        return



    def PostMeshWrapper(self, formulation, mesh, material, solver, fem_solver):
        """Calls PostMesh wrapper to get exact Dirichlet boundary conditions"""

        try:
            # from .PostMeshPy import (PostMeshCurvePy as PostMeshCurve, PostMeshSurfacePy as PostMeshSurface)
            from PostMeshPy import (PostMeshCurvePy as PostMeshCurve, PostMeshSurfacePy as PostMeshSurface)
        except ImportError:
            raise ImportError("PostMesh is not installed. Please install using 'pip install PostMeshPy'")

        from Florence.FunctionSpace import Tri

        C = mesh.InferPolynomialDegree() - 1

        if formulation.ndim == 2:

            # CHOOSE TYPE OF BOUNDARY SPACING
            boundary_fekete = np.array([[]])
            if self.nodal_spacing_for_cad == 'fekete':
                boundary_fekete = GaussLobattoQuadrature(C+2)[0]
            else:
                boundary_fekete = EquallySpacedPoints(formulation.ndim,C)
            # IT IS IMPORTANT TO ENSURE THAT THE DATA IS C-CONITGUOUS
            boundary_fekete = boundary_fekete.copy(order="c")

            curvilinear_mesh = PostMeshCurve(mesh.element_type,dimension=formulation.ndim)
            curvilinear_mesh.SetMeshElements(mesh.elements)
            curvilinear_mesh.SetMeshPoints(mesh.points)
            curvilinear_mesh.SetMeshEdges(mesh.edges)
            curvilinear_mesh.SetMeshFaces(np.zeros((1,4),dtype=np.uint64))
            curvilinear_mesh.SetScale(self.scale_value_on_projection)
            curvilinear_mesh.SetCondition(self.condition_for_projection)
            curvilinear_mesh.SetProjectionPrecision(1.0e-04)
            curvilinear_mesh.SetProjectionCriteria(self.projection_flags)
            curvilinear_mesh.ScaleMesh()
            # curvilinear_mesh.InferInterpolationPolynomialDegree()
            curvilinear_mesh.SetNodalSpacing(boundary_fekete)
            curvilinear_mesh.GetBoundaryPointsOrder()
            # READ THE GEOMETRY FROM THE IGES FILE
            curvilinear_mesh.ReadCAD(self.cad_file)
            # EXTRACT GEOMETRY INFORMATION FROM THE IGES FILE
            geometry_points = curvilinear_mesh.GetGeomVertices()
            self.GetGeometryMeshScale(geometry_points,mesh)
            # print([np.min(geometry_points[:,0]),np.max(geometry_points[:,0])], mesh.Bounds)
            # exit()
            curvilinear_mesh.GetGeomEdges()
            curvilinear_mesh.GetGeomFaces()
            curvilinear_mesh.GetGeomPointsOnCorrespondingEdges()
            # FIRST IDENTIFY WHICH CURVES CONTAIN WHICH EDGES
            curvilinear_mesh.IdentifyCurvesContainingEdges()
            # PROJECT ALL BOUNDARY POINTS FROM THE MESH TO THE CURVE
            curvilinear_mesh.ProjectMeshOnCurve()
            # FIX IMAGES AND ANTI IMAGES IN PERIODIC CURVES/SURFACES
            curvilinear_mesh.RepairDualProjectedParameters()
            # PERFORM POINT INVERSION FOR THE INTERIOR POINTS
            if self.projection_type == 'orthogonal':
                curvilinear_mesh.MeshPointInversionCurve()
            elif self.projection_type == 'arc_length':
                curvilinear_mesh.MeshPointInversionCurveArcLength()
            else:
                # warn("projection type not understood. Arc length based projection is going to be used")
                curvilinear_mesh.MeshPointInversionCurveArcLength()
            # OBTAIN MODIFIED MESH POINTS - THIS IS NECESSARY TO ENSURE LINEAR MESH IS ALSO CORRECT
            curvilinear_mesh.ReturnModifiedMeshPoints(mesh.points)
            # GET DIRICHLET DATA
            nodesDBC, Dirichlet = curvilinear_mesh.GetDirichletData()

            # GET ACTUAL CURVE POINTS - THIS FUNCTION IS EXPENSIVE
            # self.ActualCurve = curvilinear_mesh.DiscretiseCurves(100)

            if self.save_nurbs_data:
                from scipy.io import savemat
                nurbs_dict = {'nodesDBC':nodesDBC, 'Dirichlet':Dirichlet}
                savemat(self.filename, nurbs_dict, do_compression=True)

        elif formulation.ndim == 3:

            t_all_proj = time()

            boundary_points = FeketePointsTri(C)
            if mesh.element_type == "hex":
                boundary_points = GaussLobattoPointsQuad(C)

            curvilinear_mesh = PostMeshSurface(mesh.element_type,dimension=formulation.ndim)
            curvilinear_mesh.SetMeshElements(mesh.elements)
            curvilinear_mesh.SetMeshPoints(mesh.points)
            if mesh.edges is not None:
                if mesh.edges.ndim == 2 and mesh.edges.shape[1]==0:
                    mesh.edges = np.zeros((1,4),dtype=np.uint64)
                else:
                    curvilinear_mesh.SetMeshEdges(mesh.edges)
            curvilinear_mesh.SetMeshFaces(mesh.faces)
            curvilinear_mesh.SetScale(self.scale_value_on_projection)
            curvilinear_mesh.SetCondition(self.condition_for_projection)
            curvilinear_mesh.SetProjectionPrecision(1.0e-04)
            curvilinear_mesh.SetProjectionCriteria(self.projection_flags)
            curvilinear_mesh.ScaleMesh()
            curvilinear_mesh.SetNodalSpacing(boundary_points)
            # curvilinear_mesh.GetBoundaryPointsOrder()
            # READ THE GEOMETRY FROM THE IGES FILE
            curvilinear_mesh.ReadCAD(self.cad_file)
            # EXTRACT GEOMETRY INFORMATION FROM THE IGES FILE
            geometry_points = curvilinear_mesh.GetGeomVertices()
            self.GetGeometryMeshScale(geometry_points,mesh)
            # print([np.min(geometry_points[:,2]),np.max(geometry_points[:,2])], mesh.Bounds)
            # exit()
            curvilinear_mesh.GetGeomEdges()
            curvilinear_mesh.GetGeomFaces()
            print("CAD geometry has", curvilinear_mesh.NbPoints, "points,", \
            curvilinear_mesh.NbCurves, "curves and", curvilinear_mesh.NbSurfaces, "surfaces")
            curvilinear_mesh.GetGeomPointsOnCorrespondingFaces()

            # FIRST IDENTIFY WHICH SURFACES CONTAIN WHICH FACES
            if getattr(mesh,"face_to_surface",None) is not None:
                if mesh.faces.shape[0] == mesh.face_to_surface.size:
                    if mesh.face_to_surface.size != mesh.face_to_surface.shape[0]:
                        mesh.face_to_surface = np.ascontiguousarray(mesh.face_to_surface.flatten(),dtype=np.int64)
                    curvilinear_mesh.SupplySurfacesContainingFaces(mesh.face_to_surface,already_mapped=1)
                else:
                    raise AssertionError("face-to-surface mapping does not seem correct. "
                        "Point projection is going to stop")
            else:
                if self.surface_identification_algorithm == 'minimisation':
                    curvilinear_mesh.IdentifySurfacesContainingFaces(int(self.activate_bounding_box),
                        self.bounding_box_padding)
                elif self.surface_identification_algorithm == 'pure_projection':
                    curvilinear_mesh.IdentifySurfacesContainingFacesByPureProjection(int(self.activate_bounding_box),
                        self.bounding_box_padding)
                else:
                    # warn("surface identification algorithm not understood. minimisation algorithm is going to be used")
                    curvilinear_mesh.IdentifySurfacesContainingFaces(int(self.activate_bounding_box))

            if self.project_on_curves:
                t_proj = time()
                # IDENTIFY WHICH EDGES ARE SHARED BETWEEN SURFACES
                curvilinear_mesh.IdentifySurfacesIntersections()
                print("Curve intersection recognition took {} seconds".format(time()-t_proj))

            # PERFORM POINT INVERSION FOR THE INTERIOR POINTS
            if self.projection_type == "arc_length":
                assert mesh.element_type == "tet"

                Neval = np.zeros((3,boundary_points.shape[0]),dtype=np.float64)
                hpBases = Tri.hpNodal.hpBases
                for i in range(3,boundary_points.shape[0]):
                    Neval[:,i]  = hpBases(0,boundary_points[i,0],boundary_points[i,1],1)[0]

            if self.projection_type == 'orthogonal':
                curvilinear_mesh.MeshPointInversionSurface(self.project_on_curves, self.modify_linear_mesh_on_projection)
            elif self.projection_type == 'arc_length':
                # PROJECT ALL BOUNDARY POINTS FROM THE MESH TO THE SURFACE
                curvilinear_mesh.ProjectMeshOnSurface()
                # curvilinear_mesh.RepairDualProjectedParameters()
                curvilinear_mesh.MeshPointInversionSurfaceArcLength(self.project_on_curves,
                    self.orthogonal_fallback_tolerance,Neval)
            else:
                warn("projection type not understood. Orthogonal projection is going to be used")
                curvilinear_mesh.MeshPointInversionSurface(self.project_on_curves)

            # OBTAIN MODIFIED MESH POINTS - THIS IS NECESSARY TO ENSURE LINEAR MESH IS ALSO CORRECT
            if self.modify_linear_mesh_on_projection:
                curvilinear_mesh.ReturnModifiedMeshPoints(mesh.points)
            # GET DIRICHLET DATA
            nodesDBC, Dirichlet = curvilinear_mesh.GetDirichletData()
            # GET DIRICHLET FACES (IF REQUIRED)
            dirichlet_faces = curvilinear_mesh.GetDirichletFaces()

            # FOR GEOMETRIES CONTAINING PLANAR SURFACES
            planar_mesh_faces = curvilinear_mesh.GetMeshFacesOnPlanarSurfaces()
            # self.planar_mesh_faces = planar_mesh_faces

            if self.save_nurbs_data:
                from scipy.io import savemat
                nurbs_dict = {'nodesDBC':nodesDBC,
                    'Dirichlet':Dirichlet,
                    'dirichlet_faces':dirichlet_faces}
                savemat(self.filename, nurbs_dict, do_compression=True)

            print("3D multi-level projection (excluding mesh deformation) took {} seconds".format(time()-t_all_proj))

            if self.solve_for_planar_faces:
                if planar_mesh_faces.shape[0] != 0:
                    # SOLVE A 2D PROBLEM FOR PLANAR SURFACES
                    switcher = fem_solver.parallel
                    if fem_solver.parallel is True:
                        fem_solver.parallel = False

                    self.GetDirichletDataForPlanarFaces(formulation, material,
                        mesh, solver, fem_solver, planar_mesh_faces, nodesDBC, Dirichlet, plot=False)
                    fem_solver.parallel == switcher

        return nodesDBC, Dirichlet


    @staticmethod
    def GetDirichletDataForPlanarFaces(formulation, material,
        mesh, solver, fem_solver, planar_mesh_faces, nodesDBC, Dirichlet, plot=False):
        """Solve a 2D problem for planar faces. Modifies Dirichlet"""

        from copy import deepcopy
        from Florence import Mesh, FunctionSpace, QuadratureRule
        from Florence.PostProcessing import PostProcess
        from Florence.Tensor import itemfreq, makezero, in2d_unsorted

        surface_flags = itemfreq(planar_mesh_faces[:,1])
        number_of_planar_surfaces = surface_flags.shape[0]

        C = mesh.InferPolynomialDegree() - 1

        E1 = [1.,0.,0.]
        E2 = [0.,1.,0.]
        E3 = [0.,0.,1.]

        # MAKE A SINGLE INSTANCE OF MATERIAL AND UPDATE IF NECESSARY
        import Florence.MaterialLibrary
        pmaterial_func = getattr(Florence.MaterialLibrary,material.mtype,None)
        pmaterial_dict = deepcopy(material.__dict__)
        del pmaterial_dict['ndim'], pmaterial_dict['mtype']
        pmaterial = pmaterial_func(2,**pmaterial_dict)


        print("The problem requires 2D analyses. Solving", number_of_planar_surfaces, "2D problems")
        for niter in range(number_of_planar_surfaces):

            pmesh = Mesh()
            if mesh.element_type == "tet":
                pmesh.element_type = "tri"
                no_face_vertices = 3
            elif mesh.element_type == "hex":
                pmesh.element_type = "quad"
                no_face_vertices = 4
            else:
                raise ValueError("Curvilinear mesher for element type {} not yet implemented".format(mesh.element_type))

            pmesh.elements = mesh.faces[planar_mesh_faces[planar_mesh_faces[:,1]==surface_flags[niter,0],0],:]
            pmesh.nelem = np.int64(surface_flags[niter,1])
            pmesh.GetBoundaryEdges()
            unique_edges = np.unique(pmesh.edges).astype(nodesDBC.dtype)
            # Dirichlet2D = np.zeros((unique_edges.shape[0],3))
            # nodesDBC2D = np.zeros(unique_edges.shape[0]).astype(np.int64)

            unique_elements, inv  = np.unique(pmesh.elements, return_inverse=True)
            unique_elements = unique_elements.astype(nodesDBC.dtype)
            aranger = np.arange(unique_elements.shape[0],dtype=np.uint64)
            pmesh.elements = aranger[inv].reshape(pmesh.elements.shape)

            # counter = 0
            # for i in unique_edges:
            #     nodesDBC2D[counter] = np.where(nodesDBC==i)[0][0]
            #     Dirichlet2D[counter,:] = Dirichlet[nodesDBC2D[counter],:]
            #     counter += 1
            # nodesDBC2D = nodesDBC2D.astype(np.int64)

            # temp_dict = []
            # for i in nodesDBC[nodesDBC2D].flatten():
            #     temp_dict.append(np.where(unique_elements==i)[0][0])
            # nodesDBC2D = np.array(temp_dict,copy=False)
            dirichlet_edges = in2d_unsorted(nodesDBC,unique_edges[:,None]).flatten()
            nodesDBC2D = in2d_unsorted(unique_elements.astype(nodesDBC.dtype)[:,None],nodesDBC[dirichlet_edges]).flatten()
            Dirichlet2D = Dirichlet[dirichlet_edges,:]

            pmesh.points = mesh.points[unique_elements,:]

            one_element_coord = pmesh.points[pmesh.elements[0,:no_face_vertices],:]

            # FOR COORDINATE TRANSFORMATION
            AB = one_element_coord[0,:] - one_element_coord[1,:]
            AC = one_element_coord[0,:] - one_element_coord[2,:]

            normal = np.cross(AB,AC)
            unit_normal = normal/np.linalg.norm(normal)

            e1 = AB/np.linalg.norm(AB)
            e2 = np.cross(normal,AB)/np.linalg.norm(np.cross(normal,AB))
            e3 = unit_normal

            # TRANSFORMATION MATRIX
            Q = np.array([
                [np.einsum('i,i',e1,E1), np.einsum('i,i',e1,E2), np.einsum('i,i',e1,E3)],
                [np.einsum('i,i',e2,E1), np.einsum('i,i',e2,E2), np.einsum('i,i',e2,E3)],
                [np.einsum('i,i',e3,E1), np.einsum('i,i',e3,E2), np.einsum('i,i',e3,E3)]
                ])

            pmesh.points = np.dot(pmesh.points,Q.T)
            # assert np.allclose(pmesh.points[:,2],pmesh.points[0,2])
            # z_plane = pmesh.points[0,2]

            pmesh.points = pmesh.points[:,:2]

            Dirichlet2D = np.dot(Dirichlet2D,Q.T)
            Dirichlet2D = Dirichlet2D[:,:2]

            pmesh.edges = None
            pmesh.GetBoundaryEdges()

            # GET BOUNDARY CONDITION FOR 2D PROBLEM
            pboundary_condition = BoundaryCondition()
            pboundary_condition.SetCADProjectionParameters()
            pboundary_condition.is_dirichlet_computed = True
            pboundary_condition.nodesDBC = nodesDBC2D[:,None]
            pboundary_condition.Dirichlet = Dirichlet2D

            # GET VARIATIONAL FORMULATION FOR 2D PROBLEM
            # from Florence import DisplacementFormulation
            # pformulation = DisplacementFormulation(pmesh)
            pformulation_func = formulation.__class__
            pformulation = pformulation_func(pmesh)

            pfem_solver = deepcopy(fem_solver)
            pfem_solver.do_not_reset = True

            print('Solving planar problem {}. Number of DoF is {}'.format(niter,pmesh.points.shape[0]*pformulation.nvar))
            if pmesh.points.shape[0] != Dirichlet2D.shape[0]:
                # CALL THE FEM SOLVER FOR SOLVING THE 2D PROBLEM
                solution = pfem_solver.Solve(formulation=pformulation,
                    mesh=pmesh, material=pmaterial,
                    boundary_condition=pboundary_condition)
                TotalDisp = solution.sol
            else:
                # IF THERE IS NO DEGREE OF FREEDOM TO SOLVE FOR (ONE ELEMENT CASE)
                TotalDisp = Dirichlet2D[:,:,None]

            Disp = np.zeros((TotalDisp.shape[0],3))
            Disp[:,:2] = TotalDisp[:,:,-1]

            # temp_dict = []
            # for i in unique_elements:
            #     temp_dict.append(np.where(nodesDBC==i)[0][0])
            temp_dict = in2d_unsorted(nodesDBC,unique_elements[:,None]).flatten()
            Dirichlet[temp_dict,:] = np.dot(Disp,Q)

            if plot:
                post_process = PostProcess(2,2)
                post_process.CurvilinearPlot(pmesh, TotalDisp,
                    QuantityToPlot=solution.ScaledJacobian, interpolation_degree=40)
                import matplotlib.pyplot as plt
                plt.show()


            # import matplotlib.pyplot as plt
            # figure = plt.figure(figsize=(8, 8))
            # TotalDisp = np.zeros_like(TotalDisp)
            # TotalDisp[pboundary_condition.nodesDBC,:,-1] = Dirichlet2D[:,:,None].reshape(36,1,2)
            # figure=None
            # post_process = PostProcess(2,2)
            # post_process.CurvilinearPlot(pmesh, TotalDisp, interpolation_degree=40,
            # figure=figure, color="#E3A933", plot_points=True, point_radius=2.0)
            # plt.show()

            del pmesh, pboundary_condition

        gc.collect()






    def GetReducedMatrices(self,stiffness,F,mass=None):

        # GET REDUCED FORCE VECTOR
        F_b = F[self.columns_in,0]

        # GET REDUCED STIFFNESS MATRIX
        stiffness_b = stiffness[self.columns_in,:][:,self.columns_in]

        # GET REDUCED MASS MATRIX
        mass_b = np.array([])
        # if self.analysis_type != 'static':
        #     mass_b = mass[self.columns_in,:][:,self.columns_in]

        return stiffness_b, F_b, mass_b


    def ApplyDirichletGetReducedMatrices(self, stiffness, F, AppliedDirichlet, LoadFactor=1., mass=None, only_residual=False):
        """AppliedDirichlet is a non-member because it can be external incremental Dirichlet,
            which is currently not implemented as member of BoundaryCondition. F also does not
            correspond to Dirichlet forces, as it can be residual in incrementally linearised
            framework.
        """

        # # APPLY DIRICHLET BOUNDARY CONDITIONS
        # for i in range(self.columns_out.shape[0]):
            # F = F - LoadFactor*AppliedDirichlet[i]*stiffness.getcol(self.columns_out[i])

        # MUCH FASTER APPROACH
        # F = F - (stiffness[:,self.columns_out]*AppliedDirichlet*LoadFactor)[:,None]
        nnz_cols = ~np.isclose(AppliedDirichlet,0.0)
        F[self.columns_in] = F[self.columns_in] - (stiffness[self.columns_in,:][:,
            self.columns_out[nnz_cols]]*AppliedDirichlet[nnz_cols]*LoadFactor)[:,None]

        if only_residual:
            return F

        # GET REDUCED FORCE VECTOR
        F_b = F[self.columns_in,0]

        # # FOR UMFPACK SOLVER TAKE SPECIAL CARE
        # if int(sp.__version__.split('.')[1]) < 15:
        #     F_b_umf = np.zeros(F_b.shape[0])
        #     # F_b_umf[:] = F_b[:,0] # DOESN'T WORK
        #     for i in range(F_b_umf.shape[0]):
        #         # F_b_umf[i] = F_b[i,0]
        #         F_b_umf[i] = F_b.flatten()[i]
        #     F_b = np.copy(F_b_umf)


        # GET REDUCED STIFFNESS
        stiffness_b = stiffness[self.columns_in,:][:,self.columns_in]

        # GET REDUCED MASS MATRIX
        if self.analysis_type != 'static':
            mass_b = mass[self.columns_in,:][:,self.columns_in]
            return stiffness_b, F_b, F, mass_b

        return stiffness_b, F_b, F


    def GetReducedVectors(self, F, mass=None, only_residual=False):

        # GET REDUCED FORCE VECTOR
        F_b = F[self.columns_in,0]

        # GET REDUCED MASS MATRIX
        mass_b = []
        if self.analysis_type != 'static' and not only_residual:
            mass_b = mass[self.columns_in,0]

        return F_b, mass_b



    def UpdateFixDoFs(self, AppliedDirichletInc, fsize, nvar):
        """Updates the geometry (DoFs) with incremental Dirichlet boundary conditions
            for fixed/constrained degrees of freedom only. Needs to be applied per time steps"""

        # GET TOTAL SOLUTION
        TotalSol = np.zeros((fsize,1))
        TotalSol[self.columns_out,0] = AppliedDirichletInc

        # RE-ORDER SOLUTION COMPONENTS
        dU = TotalSol.reshape(int(TotalSol.shape[0]/nvar),nvar)

        return dU

    def UpdateFreeDoFs(self, sol, fsize, nvar):
        """Updates the geometry with iterative solutions of Newton-Raphson
            for free degrees of freedom only. Needs to be applied per time NR iteration"""

        # GET TOTAL SOLUTION
        TotalSol = np.zeros((fsize,1))
        TotalSol[self.columns_in,0] = sol

        # RE-ORDER SOLUTION COMPONENTS
        dU = TotalSol.reshape(int(TotalSol.shape[0]/nvar),nvar)

        return dU



    def SetNURBSParameterisation(self,nurbs_func,*args):
        self.nurbs_info = nurbs_func(*args)


    def SetNURBSCondition(self,nurbs_func,*args):
        self.nurbs_condition = nurbs_func(*args)

        # dynamic_step=0
    def ComputeNeumannForces(self, mesh, material, function_spaces, compute_traction_forces=True, compute_body_forces=False):
        """Compute/assemble traction and body forces"""

        if self.neumann_flags is None:
            return np.zeros((mesh.points.shape[0]*material.nvar,1),dtype=np.float64)

        nvar = material.nvar
        ndim = mesh.InferSpatialDimension()

        if self.neumann_flags.shape[0] == mesh.points.shape[0]:
            self.neumann_data_applied_at = "node"
        else:
            if ndim==3:
                if self.neumann_flags.shape[0] == mesh.faces.shape[0]:
                    self.neumann_data_applied_at = "face"
            elif ndim==2:
                if self.neumann_flags.shape[0] == mesh.edges.shape[0]:
                    self.neumann_data_applied_at = "face"


        if self.neumann_data_applied_at == 'face':
            from Florence.FiniteElements.Assembly import AssembleForces
            if not isinstance(function_spaces,tuple):
                raise ValueError("Boundary functional spaces not available for computing Neumman and body forces")
            else:
                # CHECK IF A FUNCTION SPACE FOR BOUNDARY EXISTS - SAFEGAURDS AGAINST FORMULATIONS THAT DO NO PROVIDE ONE
                has_boundary_spaces = False
                for fs in function_spaces:
                    if ndim == 3 and fs.ndim == 2:
                        has_boundary_spaces = True
                        break
                    elif ndim == 2 and fs.ndim == 1:
                        has_boundary_spaces = True
                        break
                if not has_boundary_spaces:
                    from Florence import QuadratureRule, FunctionSpace
                    # COMPUTE BOUNDARY FUNCTIONAL SPACES
                    p = mesh.InferPolynomialDegree()
                    bquadrature = QuadratureRule(optimal=3, norder=2*p+1,
                        mesh_type=mesh.boundary_element_type, is_flattened=False)
                    bfunction_space = FunctionSpace(mesh.CreateDummyLowerDimensionalMesh(),
                        bquadrature, p=p, equally_spaced=mesh.IsEquallySpaced, use_optimal_quadrature=False)
                    function_spaces = (function_spaces[0],bfunction_space)
                    # raise ValueError("Boundary functional spaces not available for computing Neumman and body forces")

            t_tassembly = time()
            if self.analysis_type == "static":
                F = AssembleForces(self, mesh, material, function_spaces,
                    compute_traction_forces=compute_traction_forces, compute_body_forces=compute_body_forces)
            elif self.analysis_type == "dynamic":
                if self.neumann_flags.ndim==2:
                    # THE POSITION OF NEUMANN DATA APPLIED AT FACES CAN CHANGE DYNAMICALLY
                    tmp_flags = np.copy(self.neumann_flags)
                    tmp_data = np.copy(self.applied_neumann)
                    F = np.zeros((mesh.points.shape[0]*nvar,self.neumann_flags.shape[1]))
                    for step in range(self.neumann_flags.shape[1]):
                        self.neumann_flags = tmp_flags[:,step]
                        self.applied_neumann = tmp_data[:,:,step]
                        F[:,step] = AssembleForces(self, mesh, material, function_spaces,
                            compute_traction_forces=compute_traction_forces, compute_body_forces=compute_body_forces).flatten()

                    self.neumann_flags = tmp_flags
                    self.applied_neumann = tmp_data
                else:
                    # THE POSITION OF NEUMANN DATA APPLIED AT FACES CAN CHANGE DYNAMICALLY
                    F = AssembleForces(self, mesh, material, function_spaces,
                            compute_traction_forces=compute_traction_forces, compute_body_forces=compute_body_forces).flatten()

            print("Assembled external traction forces. Time elapsed is {} seconds".format(time()-t_tassembly))


        elif self.neumann_data_applied_at == 'node':
            # A DIRICHLET TYPE METHODOLGY FOR APPLYING NEUMANN BOUNDARY CONDITONS (i.e. AT NODES)
            if self.analysis_type == "dynamic":
                if self.neumann_flags.ndim ==3:
                    # FOR DYNAMIC ANALYSIS IT IS ASSUMED THAT
                    # to_apply DOOES NOT CHANGE DURING THE ANALYSIS
                    flat_neu = self.neumann_flags[:,:,0].ravel()
                    to_apply = np.arange(self.neumann_flags[:,:,0].size)[~np.isnan(flat_neu)]
                    F = np.zeros((mesh.points.shape[0]*nvar,self.neumann_flags.shape[2]))

                    for step in range(self.neumann_flags.shape[2]):
                        flat_neu = self.neumann_flags[:,:,step].ravel()
                        to_apply = np.arange(self.neumann_flags[:,:,step].size)[~np.isnan(flat_neu)]
                        F[to_apply,step] = flat_neu[~np.isnan(flat_neu)]
                else:
                    F = np.zeros((mesh.points.shape[0]*nvar,1))
                    flat_neu = self.neumann_flags.ravel()
                    to_apply = np.arange(self.neumann_flags.size)[~np.isnan(flat_neu)]
                    applied_neumann = flat_neu[~np.isnan(flat_neu)]
                    F[to_apply,0] = applied_neumann
            else:
                F = np.zeros((mesh.points.shape[0]*nvar,1))
                flat_neu = self.neumann_flags.ravel()
                to_apply = np.arange(self.neumann_flags.size)[~np.isnan(flat_neu)]
                applied_neumann = flat_neu[~np.isnan(flat_neu)]
                F[to_apply,0] = applied_neumann

        return F




    def __dirichlet_helper__(self,stiffness, AppliedDirichlet, columns_out):
        from scipy.sparse import csc_matrix
        M = csc_matrix((AppliedDirichlet,
            (columns_out,np.zeros_like(columns_out))),
            shape=(stiffness.shape[1],1))
        return (stiffness*M).A


import numpy as np
from Florence.MeshGeneration.NodeArrangement import NodeArrangementQuad, NodeArrangementHex

def EquallySpacedPoints(ndim=2,C=0):
    """Produce equally spaced points in (ndim-1) dimension, for the boundaries
        of the mesh. For ndim=2 this is the region enclosed in [-1,1] range

        input:
            C:                  [int] order of polynomial interpolation
            Returns:            [ndarray] array of equally spaced points

        """

    if ndim==2:
        # 1D: FOR 2-DIMENSION BOUNDARIES
        return np.linspace(-1,1,C+2)[:,None]
    elif ndim==3:
        # 2D: FOR 3-DIMENSION BOUNDARIES
        xs = np.linspace(-1,1,C+2)[:,None]
        x,y = np.meshgrid(xs,xs)
        points = np.concatenate((x.flatten()[:,None],y.flatten()[:,None]),axis=1)
        node_aranger = NodeArrangementQuad(C)[2]
        return points[node_aranger,:]
    elif ndim==4:
        # 3D: ACTUAL 3D ELEMENTS
        xs = np.linspace(-1,1,C+2)[:,None]
        x,y,z = np.meshgrid(xs,xs,xs)
        points = np.concatenate((y.T.flatten()[:,None],x.T.flatten()[:,None],z.T.flatten()[:,None]),axis=1)
        node_aranger = NodeArrangementHex(C)[2]
        return points[node_aranger,:]


def EquallySpacedPointsTri(C):

    h0 = 2./(C+1)

    nodes = np.array([
        [-1.,-1.],
        [1.,-1.],
        [-1.,1.]
        ])

    for i in range(C):
        nodes = np.concatenate((nodes,[[-1.+(i+1.)*h0,-1.]]),axis=0)
    for j in range(C):
        for i in range(C+1-j):
            nodes = np.concatenate((nodes,[[-1.+i*h0,-1.+(j+1.)*h0]]),axis=0)

    return nodes


def EquallySpacedPointsTet(C, coordinates=None):
    """
        coordinates:            [ndarray] vertices of a tetrahedron
    """

    if coordinates is not None:
        xv = coordinates
    else:
        xv = np.array([
            [-1.,-1.,-1.],
            [ 1.,-1.,-1.],
            [-1., 1.,-1.],
            [-1.,-1., 1.],
            ])

    if C==0:
        return xv

    n = C+1
    p = 0

    nsize = int((n+1)*(n+2)*(n+3)/6)
    # xg = np.zeros((nsize,3))

    # for i in range ( 0, n + 1 ):
    #     for j in range ( 0, n + 1 - i ):
    #         for k in range ( 0, n + 1 - i - j ):
    #             l = n - i - j - k
    #             xg[p,0] = (i * xv[0,0] + j * xv[1,0] + k * xv[2,0] + l * xv[3,0]) / n
    #             xg[p,1] = (i * xv[0,1] + j * xv[1,1] + k * xv[2,1] + l * xv[3,1]) / n
    #             xg[p,2] = (i * xv[0,2] + j * xv[1,2] + k * xv[2,2] + l * xv[3,2]) / n
    #             # xg[p] = (i * xv[0] + j * xv[1] + k * xv[2] + l * xv[3]) / n
    #             p = p + 1

    xg = np.empty((nsize,4))
    for i in range ( 0, n + 1 ):
        for j in range ( 0, n + 1 - i ):
            for k in range ( 0, n + 1 - i - j ):
                l = n - i - j - k
                xg[p,0] = i
                xg[p,1] = j
                xg[p,2] = k
                xg[p,3] = l
                p = p + 1

    xg = (xg[:,:,None] * xv).sum(1) / n


    # xv = np.array([
    #     [-1.,-1.,-1.],
    #     [ 1.,-1.,-1.],
    #     [-1., 1.,-1.],
    #     [-1.,-1., 1.],
    #     ])

    # # spanning arrays of a 3d grid according to range(0,n+1)
    # ii,jj,kk = np.ogrid[:n+1,:n+1,:n+1]
    # # indices of the triples which fall inside the original for loop
    # inds = (jj < n+1-ii) & (kk < n+1-ii-jj)
    # # the [i,j,k] indices of the points that fall inside the for loop, in the same order
    # combs = np.vstack(np.where(inds)).T
    # # combs is now an (nsize,3)-shaped array

    # # compute "l" column too
    # lcol = n - combs.sum(axis=1)
    # combs = np.hstack((combs,lcol[:,None]))
    # # combs is now an (nsize,4)-shaped array

    # # all we need to do now is to take the matrix product of combs and xv, divide by n in the end
    # xg = np.matmul(combs,xv)/n


    # Sort accordingly
    xg = np.flipud(xg)
    msize = int((n)*(n+1)*(n+2)/6)
    tsize = int((n)*(n+1)/2)
    ind_vertices = [0,msize,msize+tsize,nsize-1]
    xg_vertices = xg[ind_vertices,:]
    xg_non_vertices = np.delete(xg,ind_vertices,axis=0)
    xg_non_vertices_sort = xg_non_vertices[np.lexsort((xg_non_vertices[:,0],
        xg_non_vertices[:,1],xg_non_vertices[:,2])),:]
    xg_ = np.concatenate((xg_vertices,xg_non_vertices_sort))

    return xg_


import numpy as np
cimport numpy as np
from cython cimport boundscheck, wraparound

cdef inline void FillConstitutiveB_(double *B, const double* SpatialGradient,
                     int ndim, int nvar, int rows, int cols) nogil:
    cdef int i

    if ndim == 2:

        for i in range(rows):
            B[i*cols*nvar] = SpatialGradient[i]
            B[i*cols*nvar+cols+1] = SpatialGradient[i+rows]
            
            # SEEMS BUGGY
            # B[i*cols*nvar+(cols-1)] = SpatialGradient[i+rows]
            # B[i*cols*nvar+(2*cols-1)] = SpatialGradient[i]

            B[i*cols*nvar+2] = SpatialGradient[i+rows]
            B[i*cols*nvar+cols+2] = SpatialGradient[i]

    elif ndim == 3:

        for i in range(rows):
            B[i*cols*nvar] = SpatialGradient[i]
            B[i*cols*nvar+cols+1] = SpatialGradient[i+rows]
            B[i*cols*nvar+2*(cols+1)] = SpatialGradient[i+2*rows]

            B[i*cols*nvar+cols+5] = SpatialGradient[i+2*rows]
            B[i*cols*nvar+2*cols+5] = SpatialGradient[i+rows]

            B[i*cols*nvar+4] = SpatialGradient[i+2*rows]
            B[i*cols*nvar+2*cols+4] = SpatialGradient[i]

            B[i*cols*nvar+3] = SpatialGradient[i+rows]
            B[i*cols*nvar+cols+3] = SpatialGradient[i]

            




@boundscheck(False)
@wraparound(False)
def FillConstitutiveB(np.ndarray[double,ndim=2,mode='c'] B, 
                     np.ndarray[double,ndim=2,mode='c'] SpatialGradient,
                     int ndim, int nvar):
    
    cdef int rows = SpatialGradient.shape[1]
    cdef int cols = B.shape[1]
    

    FillConstitutiveB_(&B[0,0],&SpatialGradient[0,0],
        ndim,nvar,rows,cols)











cdef inline void FillGeometricB_(double *B,double *SpatialGradient, 
                        double *S, double *CauchyStressTensor,
                        int ndim, int nvar, int rows, int cols) nogil:
    cdef int i

    if ndim == 2:

        for i in range(rows):
            B[i*cols*nvar] = SpatialGradient[i]
            B[i*cols*nvar+1] = SpatialGradient[i+rows]
            
            B[i*cols*nvar+(cols+2)] = SpatialGradient[i]
            B[i*cols*nvar+(cols+2)+1] = SpatialGradient[i+rows]

        S[0] = CauchyStressTensor[0]
        S[1] = CauchyStressTensor[1]
        S[4] = CauchyStressTensor[2]
        S[5] = CauchyStressTensor[3]

        S[10] = CauchyStressTensor[0]
        S[11] = CauchyStressTensor[1]
        S[14] = CauchyStressTensor[2]
        S[15] = CauchyStressTensor[3]

    elif ndim == 3:

        for i in range(rows):
            B[i*cols*nvar] = SpatialGradient[i]
            B[i*cols*nvar+1] = SpatialGradient[i+rows]
            B[i*cols*nvar+2] = SpatialGradient[i+2*rows]

            B[i*cols*nvar+(cols+3)] = SpatialGradient[i]
            B[i*cols*nvar+(cols+3)+1] = SpatialGradient[i+rows]
            B[i*cols*nvar+(cols+3)+2] = SpatialGradient[i+2*rows]

            B[i*cols*nvar+2*(cols+3)] = SpatialGradient[i]
            B[i*cols*nvar+2*(cols+3)+1] = SpatialGradient[i+rows]
            B[i*cols*nvar+2*(cols+3)+2] = SpatialGradient[i+2*rows]

        S[0] = CauchyStressTensor[0]
        S[1] = CauchyStressTensor[1]
        S[2] = CauchyStressTensor[2]
        S[9] = CauchyStressTensor[3]
        S[10] = CauchyStressTensor[4]
        S[11] = CauchyStressTensor[5]
        S[18] = CauchyStressTensor[6]
        S[19] = CauchyStressTensor[7]
        S[20] = CauchyStressTensor[8]

        S[30] = CauchyStressTensor[0]
        S[31] = CauchyStressTensor[1]
        S[32] = CauchyStressTensor[2]
        S[39] = CauchyStressTensor[3]
        S[40] = CauchyStressTensor[4]
        S[41] = CauchyStressTensor[5]
        S[48] = CauchyStressTensor[6]
        S[49] = CauchyStressTensor[7]
        S[50] = CauchyStressTensor[8]

        S[60] = CauchyStressTensor[0]
        S[61] = CauchyStressTensor[1]
        S[62] = CauchyStressTensor[2]
        S[69] = CauchyStressTensor[3]
        S[70] = CauchyStressTensor[4]
        S[71] = CauchyStressTensor[5]
        S[78] = CauchyStressTensor[6]
        S[79] = CauchyStressTensor[7]
        S[80] = CauchyStressTensor[8]


@boundscheck(False)
@wraparound(False)
def FillGeometricB(np.ndarray[double,ndim=2,mode='c'] B, 
                     np.ndarray[double,ndim=2, mode='c'] SpatialGradient,
                     np.ndarray[double,ndim=2,mode='c'] S,
                     np.ndarray[double,ndim=2,mode='c'] CauchyStressTensor, int ndim, int nvar):
    
    cdef int rows = SpatialGradient.shape[1]
    cdef int cols = B.shape[1]
    
    FillGeometricB_(&B[0,0],&SpatialGradient[0,0],&S[0,0],
        &CauchyStressTensor[0,0],ndim,nvar,rows,cols)








cdef inline void GetTotalTraction_(double *TotalTraction, const double *CauchyStressTensor, int ndim):
    if ndim == 2:
        TotalTraction[0] = CauchyStressTensor[0]
        TotalTraction[1] = CauchyStressTensor[3]
        TotalTraction[2] = CauchyStressTensor[1]
    elif ndim==3:
        TotalTraction[0] = CauchyStressTensor[0]
        TotalTraction[1] = CauchyStressTensor[4]
        TotalTraction[2] = CauchyStressTensor[8]
        TotalTraction[3] = CauchyStressTensor[1]
        TotalTraction[4] = CauchyStressTensor[2]
        TotalTraction[5] = CauchyStressTensor[5]




@boundscheck(False)
@wraparound(False)
def GetTotalTraction(np.ndarray[double,ndim=2,mode='c'] CauchyStressTensor):
    cdef np.ndarray[double,ndim=2,mode='c'] TotalTraction
    cdef int ndim = CauchyStressTensor.shape[0]
    if ndim == 2:
        TotalTraction = np.zeros((3,1),dtype=np.float64)
    elif ndim == 3:
        TotalTraction = np.zeros((6,1),dtype=np.float64)
    GetTotalTraction_(&TotalTraction[0,0], &CauchyStressTensor[0,0], ndim)
    return TotalTraction


import numpy as np
from numpy.linalg import norm

__all__ = ["PointInversionIsoparametricFEM"]


def hpBasesTetSimple(eta,zeta,xi):
    # THESE BASES ARE A BIT NUMERICALLY STABLE BECAUSE THEY ARE NOT BEING ROUNDED OFF
    # LIKE THE ONES COMING FROM Jacobi - ALTHOUGH LOWERING THE TOLERANCE FOR Jacobi GIVES THE SAME
    # MOREOVER THESE ARE ONLY FOR C=0

    # from Florence.FunctionSpace.DegenerateMappings import MapXiEtaZeta2RST
    # eta, zeta, xi = MapXiEtaZeta2RST(eta,zeta,xi)
    # eta, zeta, xi = eta[0], zeta[0], xi[0]
    N = np.array([  1.-eta-zeta-xi,
                    eta,
                    zeta,
                    xi
        ])

    dN = np.array([
        [-1.,-1.,-1.],
        [1.,0.,0.],
        [0.,1.,0.],
        [0.,0.,1.],
        ])

    return N, dN


def PointInversionIsoparametricFEM(element_type, C, LagrangeElemCoords, point,
    equally_spaced=False, tolerance=1.0e-7, maxiter=20, verbose=False, use_simple_bases=False, initial_guess=None):
    """ This is the inverse isoparametric map, in that given a physical point,
        find the isoparametric coordinates, provided that the coordinates of
        physical element 'LagrangeElemCoords' is known

        input:
            point:                  [2S(3D) vector] containing X,Y,(Z) value of the physical point
        return:
            iso_parametric_point    [2S(3D) vector] containing X,Y,(Z) value of the parametric point
            convergence_info        [boolean] if the iteration converged or not
    """


    from Florence.FunctionSpace import Tri, Tet, Quad, QuadES, Hex, HexES
    from Florence.Tensor import makezero

    # INITIAL GUESS - VERY IMPORTANT
    if initial_guess is None:
        # p_isoparametric = np.zeros(LagrangeElemCoords.shape[1])
        p_isoparametric = -np.ones(LagrangeElemCoords.shape[1])
    else:
        p_isoparametric = initial_guess
    residual = 0.
    blow_up_value = 1e10
    convergence_info = False

    for niter in range(maxiter):

        # Using the current iterative solution of isoparametric coordinate evaluate bases and gradient of bases
        if element_type == "tet":
            if not use_simple_bases:
                Neval, gradient = Tet.hpNodal.hpBases(C,p_isoparametric[0],p_isoparametric[1],
                    p_isoparametric[2], False, 1, equally_spaced=equally_spaced)
            else:
                Neval, gradient = hpBasesTetSimple(p_isoparametric[0],p_isoparametric[1],p_isoparametric[2])
        elif element_type == "hex":
            if not equally_spaced:
                Neval = Hex.LagrangeGaussLobatto(C,p_isoparametric[0],p_isoparametric[1],p_isoparametric[2]).flatten()
                gradient = Hex.GradLagrangeGaussLobatto(C,p_isoparametric[0],p_isoparametric[1],p_isoparametric[2])
            else:
                Neval = HexES.Lagrange(C,p_isoparametric[0],p_isoparametric[1],p_isoparametric[2]).flatten()
                gradient = HexES.GradLagrange(C,p_isoparametric[0],p_isoparametric[1],p_isoparametric[2])
        elif element_type == "tri":
            Neval, gradient = Tri.hpNodal.hpBases(C,p_isoparametric[0],p_isoparametric[1], False, 1, equally_spaced=equally_spaced)
        elif element_type == "quad":
            if not equally_spaced:
                Neval = Quad.LagrangeGaussLobatto(C,p_isoparametric[0],p_isoparametric[1]).flatten()
                gradient = Quad.GradLagrangeGaussLobatto(C,p_isoparametric[0],p_isoparametric[1])
            else:
                Neval = QuadES.Lagrange(C,p_isoparametric[0],p_isoparametric[1]).flatten()
                gradient = QuadES.GradLagrange(C,p_isoparametric[0],p_isoparametric[1])
        makezero(np.atleast_2d(Neval))
        makezero(gradient)

        # Find the residual X - X(N) [i.e. point - FEM interpolated point]
        residual = point - np.dot(Neval,LagrangeElemCoords)
        # print(np.dot(Neval,LagrangeElemCoords))
        # Find the isoparametric (Jacobian) matrix dX/d\Xi
        ParentGradientX = np.dot(gradient.T,LagrangeElemCoords)
        # ParentGradientX = np.fliplr(ParentGradientX)
        # Solve and update incremental solution
        old_p_isoparametric = np.copy(p_isoparametric)
        # print(p_isoparametric)
        try:
            # p_isoparametric += np.dot(np.linalg.inv(ParentGradientX), residual)
            p_isoparametric += np.linalg.solve(ParentGradientX, residual)
        except:
            convergence_info = False
            break

        # CHECK IF WITHIN TOLERANCE
        if norm(p_isoparametric - old_p_isoparametric) < tolerance:
            convergence_info = True
            break
        # print(np.linalg.norm(residual))
        if norm(residual) < tolerance*1e2:
            convergence_info = True
            break
        # OTHERWISE DEAL WITH BLOW UP SITUATION
        if np.isnan(norm(residual)) > blow_up_value or norm(residual) > blow_up_value:
            break
        if np.isnan(norm(p_isoparametric)) or norm(p_isoparametric) > blow_up_value:
            break

    if verbose:
        return p_isoparametric, convergence_info
    return p_isoparametric


from __future__ import print_function
import os, platform, sys, subprocess, imp

__all__ = ['AOTConfigure','AOTClean']



def MaterialList():
    ll_materials_mech          = [  "_LinearElastic_",
                                    "_NeoHookean_",
                                    "_MooneyRivlin_",
                                    "_NearlyIncompressibleMooneyRivlin_",
                                    "_AnisotropicMooneyRivlin_1_"
                                    ]
    ll_materials_electro_mech  = [  "_IsotropicElectroMechanics_0_",
                                    "_IsotropicElectroMechanics_3_",
                                    "_SteinmannModel_",
                                    "_IsotropicElectroMechanics_101_",
                                    "_IsotropicElectroMechanics_105_",
                                    "_IsotropicElectroMechanics_106_",
                                    "_IsotropicElectroMechanics_107_",
                                    "_IsotropicElectroMechanics_108_",
                                    "_Piezoelectric_100_"
                                    ]

    # ll_materials_mech  = []
    # ll_materials_electro_mech  = ["_IsotropicElectroMechanics_108_"]
    return ll_materials_mech, ll_materials_electro_mech

def execute(_cmd):
    _process = subprocess.Popen(_cmd, shell=True)
    _process.wait()



def AOTConfigure():

    ll_materials_mech, ll_materials_electro_mech = MaterialList()


    for material in ll_materials_mech:
        material_specific_assembler = "_LowLevelAssemblyDF_" + material

        header_f = material_specific_assembler + ".h"
        cython_f = material_specific_assembler + ".pyx"

        execute("cp _LowLevelAssemblyDF_.h " + header_f)
        execute("cp _LowLevelAssemblyDF_.pyx " + cython_f)

        # Read the header file
        f = open(header_f, "r")
        contents_h = f.readlines()
        f.close()

        # Modify header file
        contents_h[0] = "#ifndef " + material_specific_assembler.upper() + "_H\n"
        contents_h[1] = "#define " + material_specific_assembler.upper() + "_H\n"
        contents_h[5] = '#include "' + material + '.h"\n'

        for counter, line in enumerate(contents_h):
            if "_GlobalAssemblyDF_" in line:
                contents_h[counter] = contents_h[counter].replace("_GlobalAssemblyDF_", "_GlobalAssemblyDF_"+material)

            if "auto mat_obj" in line:
                if material == "_NeoHookean_" or material == "_LinearElastic_":
                    contents_h[counter] = "    auto mat_obj = " + material + "<Real>(mu,lamb);\n"
                elif material == "_MooneyRivlin_":
                    contents_h[counter] = "    auto mat_obj = " + material + "<Real>(mu1,mu2,lamb);\n"
                elif material == "_NearlyIncompressibleMooneyRivlin_":
                    contents_h[counter] = "    auto mat_obj = " + material + "<Real>(mu1,mu2,lamb);\n"
                elif material == "_AnisotropicMooneyRivlin_1_":
                    contents_h[counter] = "    auto mat_obj = " + material + "<Real>(mu1,mu2,mu3,lamb);\n"

        if material == "_AnisotropicMooneyRivlin_1_":
            for counter, line in enumerate(contents_h):
                if "mat_obj.KineticMeasures" in line:
                    contents_h[counter] = "        mat_obj.KineticMeasures(stress, hessian, ndim, ngauss, F, &anisotropic_orientations[elem*ndim]);\n"

        # Turn off geometric stiffness
        if material == "_LinearElastic_":
            for counter, line in enumerate(contents_h):
                if "std::fill" in line and "geometric_stiffness" in line:
                    contents_h[counter] = ""
                # DANGEROUS - this is relative to current line - we delete 13 lines down
                if "_GeometricStiffnessFiller_" in line:
                    contents_h[counter:counter+13] = ""


        contents_h[-1] = "#endif // " + material_specific_assembler.upper() + "_H"

        # Write
        f = open(header_f, 'w')
        for item in contents_h:
            f.write(item)


        # Read the cython wrapper file
        f = open(cython_f, "r")
        contents_c = f.readlines()
        f.close()


        for counter, line in enumerate(contents_c):
            if "_LowLevelAssemblyDF_" in line:
                contents_c[counter] = contents_c[counter].replace("_LowLevelAssemblyDF_", "_LowLevelAssemblyDF_" + material)
            if "_GlobalAssemblyDF_" in line:
                contents_c[counter] = contents_c[counter].replace("_GlobalAssemblyDF_", "_GlobalAssemblyDF_" + material)

            if "mu1, mu2, lamb =" in line:
                if material == "_NeoHookean_" or material == "_LinearElastic_":
                    contents_c[counter] = "    mu, lamb = material.mu, material.lamb\n"
                elif material == "_MooneyRivlin_":
                    contents_c[counter] = "    mu1, mu2, lamb = material.mu1, material.mu2, material.lamb\n"
                elif material == "_NearlyIncompressibleMooneyRivlin_":
                    contents_c[counter] = "    mu1, mu2, lamb = material.mu1, material.mu2, material.lamb\n"
                elif material == "_AnisotropicMooneyRivlin_1_":
                    contents_c[counter] = "    mu1, mu2, mu3, lamb = material.mu1, material.mu2, material.mu3, material.lamb\n"

        # Write
        f = open(cython_f, 'w')
        for item in contents_c:
            f.write(item)





    for material in ll_materials_electro_mech:
        material_specific_assembler = "_LowLevelAssemblyDPF_" + material

        header_f = material_specific_assembler + ".h"
        cython_f = material_specific_assembler + ".pyx"

        execute("cp _LowLevelAssemblyDPF_.h " + header_f)
        execute("cp _LowLevelAssemblyDPF_.pyx " + cython_f)

        # Read the header file
        f = open(header_f, "r")
        contents_h = f.readlines()
        f.close()

        # Modify header file
        contents_h[0] = "#ifndef " + material_specific_assembler.upper() + "_H\n"
        contents_h[1] = "#define " + material_specific_assembler.upper() + "_H\n"
        contents_h[5] = '#include "' + material + '.h"\n'
        # contents_h[7] = contents_h[7].replace("_GlobalAssemblyDPF_", "_GlobalAssemblyDPF_"+material)

        for counter, line in enumerate(contents_h):
            if "_GlobalAssemblyDPF_" in line:
                contents_h[counter] = contents_h[counter].replace("_GlobalAssemblyDPF_", "_GlobalAssemblyDPF_"+material)

            if "auto mat_obj" in line:
                if material == "_IsotropicElectroMechanics_0_":
                    contents_h[counter] = "    auto mat_obj = " + material + "<Real>(mu,lamb,eps_1);\n"
                elif material == "_IsotropicElectroMechanics_3_":
                    contents_h[counter] = "    auto mat_obj = " + material + "<Real>(mu,lamb,eps_1,eps_2);\n"
                elif material == "_SteinmannModel_":
                    contents_h[counter] = "    auto mat_obj = " + material + "<Real>(mu,lamb,eps_3,eps_2,eps_1);\n"
                elif material == "_IsotropicElectroMechanics_101_":
                    contents_h[counter] = "    auto mat_obj = " + material + "<Real>(mu,lamb,eps_1);\n"
                elif material == "_IsotropicElectroMechanics_105_":
                    contents_h[counter] = "    auto mat_obj = " + material + "<Real>(mu1,mu2,lamb,eps_1,eps_2);\n"
                elif material == "_IsotropicElectroMechanics_106_":
                    contents_h[counter] = "    auto mat_obj = " + material + "<Real>(mu1,mu2,lamb,eps_1,eps_2);\n"
                elif material == "_IsotropicElectroMechanics_107_":
                    contents_h[counter] = "    auto mat_obj = " + material + "<Real>(mu1,mu2,mue,lamb,eps_1,eps_2,eps_e);\n"
                elif material == "_IsotropicElectroMechanics_108_":
                    contents_h[counter] = "    auto mat_obj = " + material + "<Real>(mu1,mu2,lamb,eps_2);\n"
                elif material == "_Piezoelectric_100_":
                    contents_h[counter] = "    auto mat_obj = " + material + "<Real>(mu1,mu2,mu3,lamb,eps_1,eps_2,eps_3);\n"


        if material == "_Piezoelectric_100_":
            for counter, line in enumerate(contents_h):
                if "mat_obj.KineticMeasures" in line:
                    contents_h[counter] = "        mat_obj.KineticMeasures(D, stress, hessian, ndim, ngauss, F, ElectricFieldx, &anisotropic_orientations[elem*ndim]);\n"

        contents_h[-1] = "#endif // " + material_specific_assembler.upper() + "_H"

        # Write
        f = open(header_f, 'w')
        for item in contents_h:
            f.write(item)


        # Read the header file
        f = open(cython_f, "r")
        contents_c = f.readlines()
        f.close()

        for counter, line in enumerate(contents_c):
            if "_LowLevelAssemblyDPF_" in line:
                contents_c[counter] = contents_c[counter].replace("_LowLevelAssemblyDPF_", "_LowLevelAssemblyDPF_" + material)
            if "_GlobalAssemblyDPF_" in line:
                contents_c[counter] = contents_c[counter].replace("_GlobalAssemblyDPF_", "_GlobalAssemblyDPF_" + material)

            if "mu, lamb, eps_1 =" in line:
                if material == "_IsotropicElectroMechanics_0_":
                    contents_c[counter] = "    mu, lamb, eps_1 = material.mu, material.lamb, material.eps_1\n"
                elif material == "_IsotropicElectroMechanics_3_":
                    contents_c[counter] = "    mu, lamb, eps_1, eps_2 = material.mu, material.lamb, material.eps_1, material.eps_2\n"
                elif material == "_SteinmannModel_":
                    contents_c[counter] = "    mu, lamb, eps_3, eps_2, eps_1 = material.mu, material.lamb, material.c1, material.c2, material.eps_1\n"
                elif material == "_IsotropicElectroMechanics_101_":
                    contents_c[counter] = "    mu, lamb, eps_1 = material.mu, material.lamb, material.eps_1\n"
                elif material == "_IsotropicElectroMechanics_105_":
                    contents_c[counter] = "    mu1, mu2, lamb, eps_1, eps_2 = material.mu1, material.mu2, material.lamb, material.eps_1, material.eps_2\n"
                elif material == "_IsotropicElectroMechanics_106_":
                    contents_c[counter] = "    mu1, mu2, lamb, eps_1, eps_2 = material.mu1, material.mu2, material.lamb, material.eps_1, material.eps_2\n"
                elif material == "_IsotropicElectroMechanics_107_":
                    contents_c[counter] = "    mu1, mu2, mue, lamb, eps_1, eps_2, eps_e = material.mu1, " +\
                        "material.mu2, material.mue, material.lamb, material.eps_1, material.eps_2, material.eps_e\n"
                elif material == "_IsotropicElectroMechanics_108_":
                    contents_c[counter] = "    mu1, mu2, lamb, eps_2 = material.mu1, material.mu2, material.lamb, material.eps_2\n"
                elif material == "_Piezoelectric_100_":
                    contents_c[counter] = "    mu1, mu2, mu3, lamb, eps_1, eps_2, eps_3 = material.mu1, " +\
                        "material.mu2, material.mu3, material.lamb, material.eps_1, material.eps_2, material.eps_3\n"

        # Write
        f = open(cython_f, 'w')
        for item in contents_c:
            f.write(item)


def AOTClean():

    ll_materials_mech, ll_materials_electro_mech = MaterialList()

    for material in ll_materials_mech:
        material_specific_assembler = "_LowLevelAssemblyDF_" + material

        header_f = material_specific_assembler + ".h"
        cython_f = material_specific_assembler + ".pyx"

        execute("rm -rf " + header_f)
        execute("rm -rf " + cython_f)


    for material in ll_materials_electro_mech:
        material_specific_assembler = "_LowLevelAssemblyDPF_" + material

        header_f = material_specific_assembler + ".h"
        cython_f = material_specific_assembler + ".pyx"

        execute("rm -rf " + header_f)
        execute("rm -rf " + cython_f)


if __name__ == "__main__":

    args = sys.argv

    _op = None

    if len(args) > 1:
        for arg in args:
            if arg == "clean" or arg == "configure":
                if _op is not None:
                    raise RuntimeError("Multiple conflicting arguments passed to setup")
                _op = arg

    if _op == "clean":
        AOTClean()
    elif _op == "configure":
        AOTConfigure()
    else:
        AOTClean()
        AOTConfigure()




from .Base import Base
from .QuadratureRules import QuadratureRule
from .FunctionSpace import FunctionSpace
from .MeshGeneration import *
from .MaterialLibrary import *
from .VariationalPrinciple import *
from .BoundaryCondition import BoundaryCondition
from .Solver import *
from .Utils import PWD, RSWD
from .PostProcessing import *

__version__ = "0.1"



from __future__ import print_function

# COMPARE STRINGS WHICH MIGHT CONTAIN UNICODES
############################################################################
def insensitive(string):
    """Given a string, returns its lower/upper case insensitive string"""
    if getattr(str,'casefold',None) is not None:
        insen = lambda str_name: str_name.casefold()
    else:
        insen = lambda str_name: str_name.upper().lower()

    return insen(string)


def par_unpickle(fem_solver, mesh, material, formulation, 
    Eulerx, Eulerp, NeumannForces, NodalForces, Residual, K,
    TotalDisp, velocities, accelrations):
    """unpickle tuple of objects"""

    import pickle, gc, time, os, shutil, errno
    from scipy.io import savemat

    # from tempfile import mkdtemp
    # tmp_dir = mkdtemp()
    
    tmp_dir = "/home/roman/tmp"
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)

    fem_solver_file = os.path.join(tmp_dir,'fem_solver')
    mesh_file = os.path.join(tmp_dir,'mesh')
    material_file = os.path.join(tmp_dir,'material')
    Eulerx_file = os.path.join(tmp_dir,'Eulerx')
    Eulerp_file = os.path.join(tmp_dir,'Eulerp')

    t_unpickle = time.time()

    # main_data = MainData
    f = file(fem_solver_file, 'wb')
    pickle.dump(fem_solver,f,pickle.HIGHEST_PROTOCOL)
    f = file(mesh_file, 'wb')
    pickle.dump(mesh,f,pickle.HIGHEST_PROTOCOL)
    f = file(material_file, 'wb')
    pickle.dump(material,f,pickle.HIGHEST_PROTOCOL)
    f = file(Eulerx_file, 'wb')
    pickle.dump(Eulerx,f,pickle.HIGHEST_PROTOCOL)
    f = file(Eulerp_file, 'wb')
    pickle.dump(Eulerp,f,pickle.HIGHEST_PROTOCOL)
    
    savemat(os.path.join(tmp_dir,'rest.mat'),
        {'C':MainData.C,'ndim':MainData.ndim,'nvar':MainData.nvar},do_compression=True)
    del MainData, mesh, material, Eulerx, TotalPot, f
    gc.collect()

    print('Time taken for unpickling was', time.time() - t_unpickle, 'seconds')

    # try:
    #     # delete directory
    #     shutil.rmtree(tmp_dir)  
    # except OSError as exc:
    #     # ENOENT - no such file or directory
    #     if exc.errno != errno.ENOENT:  
    #         raise  IOError("No directory to delete")

    # f = file(main_data_file, 'rb')
    # MainData = pickle.load(f)
    # print MainData.solver

    return tmp_dir


def par_pickle(tmp_dir):

    import pickle, os, time, shutil, errno
    from scipy.io import loadmat

    t_pickle = time.time()

    # main_data_file = tmp_dir[1]
    # mesh_file =  tmp_dir[2]
    # material_file =  tmp_dir[3]
    # Eulerx_file =  tmp_dir[4]
    # TotalPot_file =  tmp_dir[5]

    main_data_file = os.path.join(tmp_dir,'main_data')
    mesh_file = os.path.join(tmp_dir,'mesh')
    material_file = os.path.join(tmp_dir,'material')
    Eulerx_file = os.path.join(tmp_dir,'Eulerx')
    TotalPot_file = os.path.join(tmp_dir,'TotalPot')

    f = file(main_data_file, 'rb')
    MainData = pickle.load(f)
    # print MainData.Domain
    f = file(mesh_file, 'rb')
    mesh = pickle.load(f)
    f = file(material_file, 'rb')
    material = pickle.load(f)
    f = file(Eulerx_file, 'rb')
    Eulerx = pickle.load(f)
    f = file(TotalPot_file, 'rb')
    TotalPot = pickle.load(f)

    # try:
    #     # delete directory
    #     # shutil.rmtree(tmp_dir)
    #     shutil.rmtree(tmp_dir[0])  
    # except OSError as exc:
    #     # ENOENT - no such file or directory
    #     if exc.errno != errno.ENOENT:  
    #         raise  IOError("No directory to delete")
    
    print ('Time taken for pickling was', time.time() - t_pickle, 'seconds')

    Dict = loadmat(os.path.join(tmp_dir,'rest'))
    MainData.C = int(Dict['C'])
    MainData.ndim = int(Dict['ndim'])
    MainData.nvar = int(Dict['nvar'])

    return MainData, mesh, material, Eulerx, TotalPot





class constant_camera_view(object):
    """Keeps the camera view for mayavi animations fixed
    """

    def __init__(self):

        import os
        os.environ['ETS_TOOLKIT'] = 'qt4'
        from mayavi import mlab

        self.mlab = mlab

    def __enter__(self):
        mlab = self.mlab
        self.orig_no_render = mlab.gcf().scene.disable_render
        if not self.orig_no_render:
            mlab.gcf().scene.disable_render = True
        cc = mlab.gcf().scene.camera
        self.orig_pos = cc.position
        self.orig_fp = cc.focal_point
        self.orig_view_angle = cc.view_angle
        self.orig_view_up = cc.view_up
        self.orig_clipping_range = cc.clipping_range

    def __exit__(self, t, val, trace):
        mlab=self.mlab
        cc = mlab.gcf().scene.camera
        cc.position = self.orig_pos
        cc.focal_point = self.orig_fp
        cc.view_angle =  self.orig_view_angle 
        cc.view_up = self.orig_view_up
        cc.clipping_range = self.orig_clipping_range

        if not self.orig_no_render:
            mlab.gcf().scene.disable_render = False
        if t != None:
            print(t, val, trace)
            ipdb.post_mortem(trace)


import numpy as np
from numpy import einsum
from .MaterialBase import Material
from Florence.Tensor import trace, Voigt


class IsotropicElectroMechanics_2(Material):
    """Isotropic Electromechanical Model 2
                W(C,E) = W_n(C) + c1*I: (E 0 E) + c2*C: (E 0 E)  
                W_n(C) = mu/2*C:I - mu*lnJ + lamb/2*(J-1)**2
                0 - stands for outer product
    """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(IsotropicElectroMechanics_2, self).__init__(mtype, ndim, **kwargs)
        # REQUIRES SEPARATELY
        self.nvar = self.ndim+1
        self.energy_type = "enthalpy"
        self.nature = "nonlinear"
        self.fields = "electro_mechanics"

        if self.ndim == 2:
            self.H_VoigtSize = 5
        elif self.ndim == 3:
            self.H_VoigtSize = 9

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = False
        

    def Hessian(self,StrainTensors, ElectricFieldx=0, elem=0, gcounter=0):

        mu = self.mu
        lamb = self.lamb
        c1 = self.c1
        c2 = self.c2

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        E = ElectricFieldx
        # print(E.shape, b.shape, np.dot(b,ElectricFieldx).shape)
        be = np.dot(b,ElectricFieldx).reshape(self.ndim)

        # Elasticity
        C = lamb*(2.*J-1.)*einsum("ij,kl",I,I) +(mu/J - lamb*(J-1))*( einsum("ik,jl",I,I)+einsum("il,jk",I,I) )
        C_Voigt = Voigt(C,1)

        # Coupled Tensor (e - 3rd order)
        e_voigt = 2*c2/J*( einsum('ij,k',b,be) + einsum('i,jk',be,b) )
        e_voigt = Voigt(e_voigt,1)

        # Dielectric Tensor (Permittivity - 2nd order)
        # Permittivity = -2./J*np.dot((c1*I+c2*b),b)
        Permittivity = 2./J*np.dot((c1*I+c2*b),b) 

        factor = -1.
        H1 = np.concatenate((C_Voigt,factor*e_voigt),axis=1)
        H2 = np.concatenate((factor*e_voigt.T,Permittivity),axis=1)
        H_Voigt = np.concatenate((H1,H2),axis=0)

        self.H_VoigtSize = H_Voigt.shape[0]

        return H_Voigt


    def CauchyStress(self, StrainTensors, ElectricFieldx, elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb
        c2 = self.c2

        I = StrainTensors['I']
        F = StrainTensors['F'][gcounter]
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        E = ElectricFieldx

        fe = np.dot(F,ElectricFieldx)

        return 1.0*mu/J*(b - I) + lamb*(J-1)*I + 2*c2*np.dot(fe,fe.T) 


    def ElectricDisplacementx(self, StrainTensors, ElectricFieldx, elem=0, gcounter=0):
        
        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        
        c1 = self.c1
        c2 = self.c2

        return -2.0/J*np.dot(b,np.dot((c1*I+c2*b),ElectricFieldx)).reshape(self.ndim,1)
        # return -2.0*np.dot(b,np.dot((c1*I+c2*b),ElectricFieldx))



import numpy as np
from .MaterialBase import Material
from Florence.Tensor import trace

#####################################################################################################
                                        # NeoHookean Material Model 2
#####################################################################################################


class NeoHookeanCoercive(Material):
    """Material model for neo-Hookean with the following internal energy:

        W(C) = mu/2*(C:I)+mu*e**(1-J)/J+lamba/2*(J-1)**2

        """

    def __init__(self, ndim, **kwargs):
    # def __init__(self, ndim, MaterialArgs=None):
        mtype = type(self).__name__
        super(NeoHookeanCoercive, self).__init__(mtype, ndim, **kwargs)
        self.ndim = ndim
        self.nvar = self.ndim

        # self.mu = self.mu
        # self.lamb = self.lamb  - self.mu

        self.is_transversely_isotropic = False
        self.energy_type = "internal_energy"
        self.nature = "nonlinear"
        self.fields = "mechanics"

        if self.ndim==3:
            self.H_VoigtSize = 6
        elif self.ndim==2:
            self.H_VoigtSize = 3

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = False


    def Hessian(self,StrainTensors,ElectricFieldx=0,elem=0,gcounter=0):

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]

        H_Voigt = self.lamb*(2.*J-1)*self.vIijIkl - self.lamb*(J-1.)*self.vIikIjl + \
                1.0*self.mu*J*np.exp(1.-J)*self.vIijIkl + self.mu*(J+1.)/J*np.exp(1.-J)*self.vIikIjl

        self.H_VoigtSize = H_Voigt.shape[0]

        return H_Voigt

    def CauchyStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]

        # stress = 2.0*self.mu/J*b + self.lamb*(J-1.0)*I - 2.*J**(-2)*self.mu*np.exp(-J+1.)*I
        # stress = self.mu/J*b + self.lamb*(J-1.0)*I - J**(-2)*self.mu*np.exp(-J+1.)*I
        stress = 1.0*self.mu/J*b + self.lamb*(J-1.0)*I - self.mu*(J+1.)/J*np.exp(1.-J)*I

        return stress





import numpy as np
from numpy import einsum
from .MaterialBase import Material
from Florence.Tensor import trace, Voigt, UnVoigt


class BonetTranservselyIsotropicHyperElastic(Material):
    """A compressible transervely isotropic model based on Bonet 1998.
        Material model is not polyconvex
    """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(BonetTranservselyIsotropicHyperElastic, self).__init__(mtype, ndim, **kwargs)
        self.is_transversely_isotropic = True
        self.is_nonisotropic = True

        self.energy_type = "internal_energy"
        self.nature = "nonlinear"
        self.fields = "mechanics"
        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = False


    def Hessian(self,StrainTensors,ElectricFieldx=0,elem=0,gcounter=0):

        # CHECK IF FIBRE ORIENTATION IS SET
        if self.anisotropic_orientations is None:
            raise ValueError("Fibre orientation for non-isotropic material is not available")

        # Get material constants (5 in this case)
        E = self.E
        E_A = self.E_A
        G_A = self.G_A
        v = self.nu
        mu = self.mu

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        F = StrainTensors['F'][gcounter]
        H = J*np.linalg.inv(F).T
        N = self.anisotropic_orientations[elem,:,None]
        FN = np.dot(F,N)[:,0]

        alpha = E/8.0/(1.+v)
        beta = E/8.0/(1.+v)
        eta_1 = 4.*alpha - G_A
        lamb = - (3*E)/(2*(v + 1)) - (E*(- E*v**2 + E_A))/((v + 1)*(2*E*v**2 + E_A*v - E_A))
        eta_2 = E/(4*(v + 1)) - (E_A*E*v)/(4*(2*E*v**2 + E_A*v - E_A)) + \
                    (E*(- E*v**2 + E_A))/(4*(v + 1)*(2*E*v**2 + E_A*v - E_A))
        gamma = (E_A**2*(v - 1))/(8*(2*E*v**2 + E_A*v - E_A)) - G_A/2 + \
                    (E_A*E*v)/(4*(2*E*v**2 + E_A*v - E_A)) - \
                    (E*(- E*v**2 + E_A))/(8*(v + 1)*(2*E*v**2 + E_A*v - E_A))

        ut = 2*alpha + 4*beta


        H_Voigt = 2.*beta/J* ( 2.0*einsum('ij,kl',b,b) - einsum('ik,jl',b,b) - einsum('il,jk',b,b) ) + \
                lamb*(2.*J-1.) *einsum('ij,kl',I,I) + \
                (ut/J - lamb*(J-1.) ) * ( einsum('ik,jl',I,I) + einsum('il,jk',I,I) ) + \
                4.*eta_2/J*( einsum('ij,k,l',b,FN,FN) + einsum('i,j,kl',FN,FN,b)  ) + \
                8.*gamma/J*( einsum('i,j,k,l',FN,FN,FN,FN) ) - \
                eta_1/J*( einsum('jk,i,l',b,FN,FN) + einsum('ik,j,l',b,FN,FN)  +
                        einsum('jl,i,k',b,FN,FN) + einsum('il,j,k',b,FN,FN) )


        H_Voigt = Voigt(H_Voigt ,1)

        self.H_VoigtSize = H_Voigt.shape[0]

        return H_Voigt



    def CauchyStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        F = StrainTensors['F'][gcounter]
        H = J*np.linalg.inv(F).T
        # N = np.array([-1.,0.]).reshape(2,1)
        N = self.anisotropic_orientations[elem,:,None]
        FN = np.dot(F,N)
        # HN = np.dot(H,N)[:,0]
        innerFN = np.dot(FN.T,FN)[0][0]
        outerFN = np.dot(FN,FN.T)
        # innerFN = einsum('i,i',FN,FN)
        # outerFN = einsum('i,j',FN,FN)
        bFN = np.dot(b,FN)

        E = self.E
        E_A = self.E_A
        G_A = self.G_A
        v = self.nu
        mu = self.mu

        alpha = E/8.0/(1.+v)
        beta = E/8.0/(1.+v)
        eta_1 = 4.*alpha - G_A
        lamb = - (3*E)/(2*(v + 1)) - (E*(- E*v**2 + E_A))/((v + 1)*(2*E*v**2 + E_A*v - E_A))
        eta_2 = E/(4*(v + 1)) - (E_A*E*v)/(4*(2*E*v**2 + E_A*v - E_A)) + \
                    (E*(- E*v**2 + E_A))/(4*(v + 1)*(2*E*v**2 + E_A*v - E_A))
        gamma = (E_A**2*(v - 1))/(8*(2*E*v**2 + E_A*v - E_A)) - G_A/2 + \
                    (E_A*E*v)/(4*(2*E*v**2 + E_A*v - E_A)) - \
                    (E*(- E*v**2 + E_A))/(8*(v + 1)*(2*E*v**2 + E_A*v - E_A))

        ut = 2*alpha + 4*beta



        if self.ndim == 3:
            trb = trace(b)
        elif self.ndim == 2:
            trb = trace(b) + 1


        stress = 2.*alpha/J*b + 2.*beta/J*(trb*b - np.dot(b,b)) - ut/J*I + lamb*(J-1.)*I + \
            2.*eta_1/J*outerFN + 2.*eta_2/J*(innerFN-1.)*b + 2.*eta_2/J*(trb-3.)*outerFN + \
            4.*gamma/J*(innerFN-1.)*outerFN - eta_1/J *(np.dot(bFN,FN.T)+np.dot(FN,bFN.T))



        return stress
